#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 17 08:55:47 2017

@author: marta
"""
from __future__ import print_function, division
from .io import load_lcurve
from .io import save_as_qdp
from .base import hen_root
import logging
from stingray.utils import excess_variance


def fvar(lc):
    return excess_variance(lc, normalization='fvar')


def excvar_none(lc):
    return excess_variance(lc, normalization='none')


def excvar_norm(lc):
    return excess_variance(lc, normalization='norm_xs')


def main(args=None):
    import argparse
    description = ('Calculate excess variance in light curve chunks')
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of files", nargs='+')
    parser.add_argument('-c', "--chunk-length", type=float, default=20,
                        help="Length in seconds of the light curve chunks")
    parser.add_argument("--fraction-step", type=float, default=0.5,
                        help="If the step is not a full chunk_length but less,"
                        "this indicates the ratio between step step and"
                        " `chunk_length`")
    parser.add_argument("--norm", type=str, default="excvar",
                        help="Choose between fvar, excvar and norm_excvar "
                             "normalization, referring to Fvar, excess "
                             "variance, and normalized excess variance "
                             "respectively (see"
                             " Vaughan et al. 2003 for details).")
    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING', type=str)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')
    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = 'DEBUG'

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENexcvar.log', level=numeric_level,
                        filemode='w')
    filelist = []
    for fname in args.files:
        lcurve = load_lcurve(fname)
        if args.norm == "fvar":
            start, stop, res = \
                lcurve.analyze_lc_chunks(args.chunk_length, fvar,
                                         args.fraction_step)
        elif args.norm == "excvar":
            start, stop, res = \
                lcurve.analyze_lc_chunks(args.chunk_length, excvar_none,
                                         args.fraction_step)
        elif args.norm == "norm_excvar":
            start, stop, res = lcurve.analyze_lc_chunks(args.chunk_length,
                                                        excvar_norm,
                                                        args.fraction_step)
        else:
            raise ValueError("Normalization must be fvar, norm_excvar "
                             "or excvar")
        var, var_err = res
        out = hen_root(fname) + "_" + args.norm + '.qdp'
        save_as_qdp([(start+stop)/2, var], [(stop-start)/2, var_err],
                    filename=out)
        filelist.append(out)

    return filelist



# -*- coding: utf-8 -*-
from __future__ import print_function, division
from .io import load_pds, HEN_FILE_EXTENSION
from .io import save_as_qdp
import logging
from .base import hen_root


def main(args=None):
    import argparse
    description = ('Read timelags from cross spectrum results and save them'
                   ' to a qdp file')
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of files", nargs='+')

    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING', type=str)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')
    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = 'DEBUG'

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENlags.log', level=numeric_level,
                        filemode='w')
    filelist = []
    for fname in args.files:
        cross = load_pds(fname)

        lag, lag_err = cross.time_lag()
        out = hen_root(fname) + '_lags.qdp'
        save_as_qdp([cross.freq, lag], [None, lag_err], filename=out)
        filelist.append(out)

    return filelist



"""Interactive phaseogram."""

from __future__ import (absolute_import, unicode_literals, division,
                        print_function)

from .io import load_events, load_folding
from .fold import get_TOAs_from_events
from .base import hen_root
from stingray.pulse.search import phaseogram
from stingray.utils import assign_value_if_none

import numpy as np
import logging
import argparse
import matplotlib.pyplot as plt
import six
from abc import ABCMeta, abstractmethod
from matplotlib.widgets import Slider, Button
import warnings


class SliderOnSteroids(Slider):
    def __init__(self, *args, **kwargs):
        self.hardvalmin = None
        self.hardvalmax = None
        if 'hardvalmin' in kwargs:
            self.hardvalmin = kwargs['hardvalmin']
            kwargs.pop('hardvalmin')
        if 'hardvalmax' in kwargs:
            self.hardvalmax = kwargs['hardvalmax']
            kwargs.pop('hardvalmax')
        Slider.__init__(self, *args, **kwargs)


@six.add_metaclass(ABCMeta)
class BasePhaseogram(object):
    def __init__(self, ev_times, freq, nph=128, nt=128, test=False,
                 fdot=0, fddot=0, mjdref=None, pepoch=None, gti=None,
                 label="phaseogram", **kwargs):
        """Init BasePhaseogram class.

        Parameters
        ----------
        ev_times : array-like
            Event times
        freq : float
            Frequency of pulsation

        Other parameters
        ----------------
        nph : int
            Number of phase bins in the profile
        nt : int
            Number of time bins in the profile
        pepoch : float, default None
            Epoch of timing solution, in the same units as ev_times
        mjdref : float, default None
            Reference MJD
        fdot : float
            First frequency derivative
        fddot : float
            Second frequency derivative
        **kwargs : keyword args
            additional arguments to pass to `self._construct_widgets`
        """

        self.fdot = fdot
        self.fddot = fddot
        self.nt = nt
        self.nph = nph
        self.mjdref = mjdref
        self.gti = gti
        self.label = label
        self.test = test

        self.pepoch = assign_value_if_none(pepoch, ev_times[0])
        self.ev_times = ev_times
        self.freq = freq

        self.fig, ax = plt.subplots()
        plt.subplots_adjust(left=0.25, bottom=0.30)

        corrected_times = self.ev_times - self._delay_fun(self.ev_times)
        self.phaseogr, phases, times, additional_info = \
            phaseogram(corrected_times, freq, return_plot=True, nph=nph, nt=nt,
                       fdot=fdot, fddot=fddot, plot=False, pepoch=pepoch)
        self.phases, self.times = phases, times

        self.pcolor = plt.pcolormesh(phases, times, self.phaseogr.T,
                                     cmap='magma')
        plt.xlabel('Phase')
        plt.ylabel('Time')
        plt.colorbar()
        self.lines = []
        self.line_phases = np.arange(-2, 3, 0.5)
        for ph0 in self.line_phases:
            newline, = plt.plot(np.zeros_like(times) + ph0, times, zorder=10,
                                lw=2, color='w')
            self.lines.append(newline)

        plt.xlim([0, 2])

        axcolor = '#ff8888'
        self.slider_axes = []
        self.slider_axes.append(plt.axes([0.25, 0.1, 0.5, 0.03],
                                         facecolor=axcolor))
        self.slider_axes.append(plt.axes([0.25, 0.15, 0.5, 0.03],
                                         facecolor=axcolor))
        self.slider_axes.append(plt.axes([0.25, 0.2, 0.5, 0.03],
                                         facecolor=axcolor))

        self._construct_widgets(**kwargs)

        self.closeax = plt.axes([0.15, 0.020, 0.15, 0.04])
        self.button_close = Button(self.closeax, 'Quit', color=axcolor,
                                   hovercolor='0.8')

        self.recalcax = plt.axes([0.3, 0.020, 0.15, 0.04])
        self.button_recalc = Button(self.recalcax, 'Recalculate',
                                    color=axcolor,
                                    hovercolor='0.975')

        self.resetax = plt.axes([0.45, 0.020, 0.15, 0.04])
        self.button_reset = Button(self.resetax, 'Reset', color=axcolor,
                                   hovercolor='0.975')

        self.zoominax = plt.axes([0.6, 0.020, 0.1, 0.04])
        self.button_zoomin = Button(self.zoominax, '+Zoom', color=axcolor,
                                    hovercolor='0.975')

        self.zoomoutax = plt.axes([0.7, 0.020, 0.1, 0.04])
        self.button_zoomout = Button(self.zoomoutax, '-Zoom', color=axcolor,
                                     hovercolor='0.975')

        self.toaax = plt.axes([0.8, 0.020, 0.1, 0.04])
        self.button_toa = Button(self.toaax, 'TOA', color=axcolor,
                                 hovercolor='0.975')

        self.button_reset.on_clicked(self.reset)
        self.button_zoomin.on_clicked(self.zoom_in)
        self.button_zoomout.on_clicked(self.zoom_out)
        self.button_toa.on_clicked(self.toa)
        self.button_recalc.on_clicked(self.recalculate)
        self.button_close.on_clicked(self.quit)

        if not test:
            plt.show()

    @abstractmethod
    def _construct_widgets(self, **kwargs):  # pragma: no cover
        pass

    @abstractmethod
    def update(self, val):  # pragma: no cover
        pass

    @abstractmethod
    def recalculate(self, event):  # pragma: no cover
        pass

    def toa(self, event):  # pragma: no cover
        warnings.warn("This function was not implemented for this Phaseogram. "
                      "Try the basic one.")

    def reset(self, event):
        for s in self.sliders:
            s.reset()
        self.pcolor.set_array(self.phaseogr.T.ravel())
        self._set_lines(False)

    def zoom_in(self, event):
        for s in self.sliders:
            valinit = s.val
            valrange = s.valmax - s.valmin
            valmin = valinit - valrange / 4
            if s.hardvalmin is not None:
                valmin = max(s.hardvalmin, valmin)
            valmax = valinit + valrange / 4
            if s.hardvalmax is not None:
                valmax = min(s.hardvalmax, valmax)
            label = s.label.get_text()
            ax = s.ax
            ax.clear()

            s.__init__(ax, label, valmin=valmin, valmax=valmax,
                       valinit=valinit, hardvalmax=s.hardvalmax,
                       hardvalmin=s.hardvalmin)
            ax.text(0, 0, str(valmin), transform=ax.transAxes,
                    horizontalalignment='left', color='white')
            ax.text(1, 0, str(valmax), transform=ax.transAxes,
                    horizontalalignment='right', color='white')
            s.on_changed(self.update)

    def zoom_out(self, event):
        for s in self.sliders:
            valinit = s.val
            valrange = s.valmax - s.valmin
            valmin = valinit - valrange
            if s.hardvalmin is not None:
                valmin = max(s.hardvalmin, valmin)
            valmax = valinit + valrange
            if s.hardvalmax is not None:
                valmax = min(s.hardvalmax, valmax)
            label = s.label.get_text()
            ax = s.ax
            ax.clear()

            s.__init__(ax, label, valmin=valmin, valmax=valmax,
                       valinit=valinit, hardvalmax=s.hardvalmax,
                       hardvalmin=s.hardvalmin)
            ax.text(0, 0, str(valmin), transform=ax.transAxes,
                    horizontalalignment='left', color='white')
            ax.text(1, 0, str(valmax), transform=ax.transAxes,
                    horizontalalignment='right', color='white')
            s.on_changed(self.update)

    @abstractmethod
    def quit(self, event):  # pragma: no cover
        pass

    @abstractmethod
    def get_values(self):  # pragma: no cover
        pass

    @abstractmethod
    def _line_delay_fun(self, times):  # pragma: no cover
        pass

    @abstractmethod
    def _delay_fun(self, times):  # pragma: no cover
        """This is the delay function _without_ frequency derivatives."""
        pass

    @abstractmethod
    def _read_sliders(self):  # pragma: no cover
        pass

    def _set_lines(self, apply_delay=True):
        if apply_delay:
            func = self._line_delay_fun
        else:
            def func(x): return 0

        for i, ph0 in enumerate(self.line_phases):
            linephase = ph0 + func(self.times) - func(self.times[0])
            self.lines[i].set_xdata(linephase)


class InteractivePhaseogram(BasePhaseogram):

    def _construct_widgets(self):
        self.df = 0
        self.dfdot = 0
        self.dfddot = 0

        tseg = np.median(np.diff(self.times))
        tobs = tseg * self.nt
        delta_df_start = 4 / tobs
        self.df_order_of_mag = np.int(np.log10(delta_df_start))
        delta_df = delta_df_start / 10 ** self.df_order_of_mag

        delta_dfdot_start = 8 / tobs ** 2
        self.dfdot_order_of_mag = np.int(np.log10(delta_dfdot_start))
        delta_dfdot = delta_dfdot_start / 10 ** self.dfdot_order_of_mag

        delta_dfddot_start = 16 / tobs ** 3
        self.dfddot_order_of_mag = np.int(np.log10(delta_dfddot_start))
        delta_dfddot = delta_dfddot_start / 10 ** self.dfddot_order_of_mag

        self.sfreq = \
            SliderOnSteroids(
                self.slider_axes[0],
                'Delta freq x$10^{}$'.format(self.df_order_of_mag),
                -delta_df, delta_df, valinit=self.df,
                hardvalmin=0)
        self.sfdot = \
            SliderOnSteroids(
                self.slider_axes[1],
                'Delta fdot x$10^{}$'.format(self.dfdot_order_of_mag),
                -delta_dfdot, delta_dfdot, valinit=self.dfdot)

        self.sfddot = \
            SliderOnSteroids(
                self.slider_axes[2],
                'Delta fddot x$10^{}$'.format(self.dfddot_order_of_mag),
                -delta_dfddot, delta_dfddot, valinit=self.dfddot)

        self.sfreq.on_changed(self.update)
        self.sfdot.on_changed(self.update)
        self.sfddot.on_changed(self.update)
        self.sliders = [self.sfreq, self.sfdot, self.sfddot]

    def update(self, val):
        self._set_lines()
        self.fig.canvas.draw_idle()

    def _read_sliders(self):
        fddot = self.sfddot.val * 10 ** self.dfddot_order_of_mag
        fdot = self.sfdot.val * 10 ** self.dfdot_order_of_mag
        freq = self.sfreq.val * 10 ** self.df_order_of_mag
        return freq, fdot, fddot

    def _line_delay_fun(self, times):
        freq, fdot, fddot = self._read_sliders()
        return ((times - self.pepoch).astype(np.float64) * freq +
                0.5 * (times - self.pepoch) ** 2 * fdot +
                1/6 * (times - self.pepoch) ** 3 * fddot)

    def _delay_fun(self, times):
        """This is the delay function _without_ frequency derivatives."""
        return 0

    def recalculate(self, event):
        dfreq, dfdot, dfddot = self._read_sliders()
        pepoch = self.pepoch

        self.fddot = self.fddot - dfddot
        self.fdot = self.fdot - dfdot
        self.freq = self.freq - dfreq

        self.phaseogr, _, _, _ = \
            phaseogram(self.ev_times, self.freq, fdot=self.fdot, plot=False,
                       nph=self.nph, nt=self.nt, pepoch=pepoch,
                       fddot=self.fddot)

        self.reset(1)

        self.fig.canvas.draw()
        print("------------------------")
        print("PEPOCH (MET)   {}".format(self.pepoch))
        if self.mjdref is not None:
            print("PEPOCH (MJD)   {}".format(self.pepoch / 86400 + self.mjdref))
        print("F0 (Hz)        {}".format(self.freq))
        print("F1 (Hz/s)      {}".format(self.fdot))
        print("F2 (Hz/s^2)    {}".format(self.fddot))
        print("------------------------")

    def toa(self, event):
        dfreq, dfdot, dfddot = self._read_sliders()
        freqs = [self.freq - dfreq, self.fdot - dfdot, self.fddot - dfddot]
        folding_length = np.median(np.diff(self.times))
        toa, toaerr = \
            get_TOAs_from_events(self.ev_times, folding_length, *freqs,
                                 gti=self.gti, template=None,
                                 mjdref=self.mjdref, nbin=self.nph,
                                 pepoch=self.pepoch,
                                 timfile=self.label + '.tim',
                                 label=self.label[:10],
                                 quick=self.test)

    def quit(self, event):
        plt.close(self.fig)

    def get_values(self):
        return self.freq, self.fdot, self.fddot


class BinaryPhaseogram(BasePhaseogram):
    def __init__(self, *args, **kwargs):
        """Init BinaryPhaseogram class.

        Parameters
        ----------
        ev_times : array-like
            Event times
        freq : float
            Frequency of pulsation

        Other parameters
        ----------------
        orbital_period : float
            orbital period in seconds
        asini : float
            projected semi-major axis in light-sec
        T0 : float
            passage through the ascending node, in seconds.
        **kwargs : keyword args
            additional arguments to pass to `BasePhaseogram`
        """
        self.orbital_period = None
        self.asini = 0
        self.t0 = None

        if 'orbital_period' in kwargs:
            self.orbital_period = kwargs['orbital_period']
            kwargs.pop('orbital_period')
        if 'asini' in kwargs:
            self.asini = kwargs['asini']
            kwargs.pop('asini')
        if 't0' in kwargs:
            self.t0 = kwargs['t0']
            kwargs.pop('t0')
        BasePhaseogram.__init__(self, *args, **kwargs)

    def _construct_widgets(self):
        self.dperiod = 0
        self.dasini = 0
        self.dt0 = 0

        tseg = np.median(np.diff(self.times))
        tobs = tseg * self.nt

        delta_period = tobs * 5
        delta_asini = 5 / self.freq
        delta_t0 = delta_period

        self.speriod = \
            SliderOnSteroids(
                self.slider_axes[0],
                'Orb. PEr. (s)',
                np.max([0, self.orbital_period - delta_period]),
                self.orbital_period + delta_period,
                valinit=self.orbital_period,
                hardvalmin=0)

        self.sasini = \
            SliderOnSteroids(
                self.slider_axes[1],
                'a sin i / c (l-sec)', 0, self.asini + delta_asini,
                valinit=self.asini,
                hardvalmin=0)

        self.st0 = \
            SliderOnSteroids(
                self.slider_axes[2],
                'T0 (MET)', self.t0 - delta_t0, self.t0 + delta_t0,
                valinit=self.t0,
                hardvalmin=0)

        self.speriod.on_changed(self.update)
        self.sasini.on_changed(self.update)
        self.st0.on_changed(self.update)
        self.sliders = [self.speriod, self.sasini, self.st0]

    def update(self, val):
        self._set_lines()
        self.fig.canvas.draw_idle()

    def _read_sliders(self):
        return self.speriod.val, self.sasini.val, self.st0.val

    def _line_delay_fun(self, times):
        orbital_period, asini, t0 = self._read_sliders()

        new_values = asini * np.sin(2 * np.pi * (times - t0) / orbital_period)
        old_values = \
            self.asini * np.sin(2 * np.pi * (times - self.t0) /
                                (self.orbital_period))
        return (new_values - old_values) * self.freq

    def _delay_fun(self, times):
        if self.t0 is None:
            self.t0 = self.pepoch
        if self.orbital_period is None:
            self.orbital_period = self.ev_times[-1] - self.ev_times[0]

        return \
            self.asini * np.sin(2 * np.pi * (times - self.t0) /
                                self.orbital_period)

    def recalculate(self, event):
        self.orbital_period, self.asini, self.t0 = self._read_sliders()

        corrected_times = self.ev_times - self._delay_fun(self.ev_times)

        self.phaseogr, _, _, _ = \
            phaseogram(corrected_times, self.freq, fdot=self.fdot, plot=False,
                       nph=self.nph, nt=self.nt, pepoch=self.pepoch,
                       fddot=self.fddot)

        self._set_lines(False)
        self.pcolor.set_array(self.phaseogr.T.ravel())
        self.sasini.valinit = self.asini
        self.speriod.valinit = self.orbital_period
        self.st0.valinit = self.t0
        self.st0.valmin = self.t0 - self.orbital_period
        self.st0.valmax = self.t0 + self.orbital_period
        self.fig.canvas.draw()
        print("------------------------")
        print("PB (s)     {}  ({} d)".format(self.orbital_period,
                                             self.orbital_period / 86400))
        print("A1 (l-s)   {}".format(self.asini))
        print("T0 (MET)   {}".format(self.t0))
        if self.mjdref is not None:
            print("T0 (MJD)   {}".format(self.t0 / 86400 + self.mjdref))
        print("------------------------")

    def quit(self, event):
        plt.close(self.fig)

    def get_values(self):
        return self.orbital_period, self.asini, self.t0


def run_interactive_phaseogram(event_file, freq, fdot=0, fddot=0, nbin=64,
                               nt=32, binary=False, test=False,
                               binary_parameters=[None, 0, None],
                               pepoch=None):
    events = load_events(event_file)
    if pepoch is None:
        pepoch = events.gti[0, 0]
    else:
        pepoch = (pepoch - events.mjdref) * 86400

    if binary:
        ip = BinaryPhaseogram(events.time, freq, nph=nbin, nt=nt,
                              fdot=fdot, test=test, fddot=fddot,
                              pepoch=pepoch,
                              orbital_period=binary_parameters[0],
                              asini=binary_parameters[1],
                              t0=binary_parameters[2],
                              mjdref=events.mjdref,
                              gti=events.gti,
                              label=hen_root(event_file))
    else:
        ip = InteractivePhaseogram(events.time, freq, nph=nbin, nt=nt,
                                   fdot=fdot, test=test, fddot=fddot,
                                   pepoch=pepoch,
                                   mjdref=events.mjdref,
                                   gti=events.gti,
                                   label=hen_root(event_file))

    return ip


def main_phaseogram(args=None):
    description = ('Plot an interactive phaseogram')
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("file", help="Input event file", type=str)
    parser.add_argument("-f", "--freq", type=float, required=False,
                        help="Initial frequency to fold", default=None)
    parser.add_argument("--fdot", type=float, required=False,
                        help="Initial fdot", default=0)
    parser.add_argument("--fddot", type=float, required=False,
                        help="Initial fddot", default=0)
    parser.add_argument("--pepoch", type=float, required=False,
                        help="Reference epoch for timing parameters",
                        default=None)
    parser.add_argument("--periodogram", type=str, required=False,
                        help="Periodogram file", default=None)
    parser.add_argument('-n', "--nbin", default=128, type=int,
                        help="Number of phase bins (X axis) of the profile")
    parser.add_argument("--ntimes", default=64, type=int,
                        help="Number of time bins (Y axis) of the phaseogram")
    parser.add_argument("--binary", help="Interact on binary parameters "
                                         "instead of frequency derivatives",
                        default=False, action='store_true')
    parser.add_argument("--binary-parameters",
                        help="Initial values for binary parameters",
                        default=[None, 0, None], nargs=3, type=float)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')
    parser.add_argument("--test",
                        help="Just a test. Destroys the window immediately",
                        default=False, action='store_true')
    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING',
                        type=str)

    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = 'DEBUG'

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENefsearch.log', level=numeric_level,
                        filemode='w')

    if args.periodogram is None and args.freq is None:
        raise ValueError('One of -f or --periodogram arguments MUST be '
                         'specified')
    elif args.periodogram is not None:
        periodogram = load_folding(args.periodogram)
        frequency = float(periodogram.peaks[0])
        fdot = 0
    else:
        frequency = args.freq
        fdot = args.fdot

    ip = run_interactive_phaseogram(args.file, freq=frequency, fdot=fdot,
                                    nbin=args.nbin, nt=args.ntimes,
                                    test=args.test, binary=args.binary,
                                    binary_parameters=args.binary_parameters,
                                    pepoch=args.pepoch)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions to calculate frequency spectra."""
from __future__ import (absolute_import, division,
                        print_function)

from .base import hen_root, common_name, _empty, _assign_value_if_none
from .io import sort_files, get_file_type, load_data, save_pds, load_lcurve
from .io import HEN_FILE_EXTENSION
from stingray.gti import cross_gtis, create_gti_mask
from stingray.crossspectrum import AveragedCrossspectrum
from stingray.powerspectrum import AveragedPowerspectrum
import numpy as np
import logging
import warnings
from multiprocessing import Pool
import os


def _wrap_fun_cpds(arglist):
    f1, f2, outname, kwargs = arglist
    return calc_cpds(f1, f2, outname=outname, **kwargs)


def _wrap_fun_pds(argdict):
    fname = argdict["fname"]
    argdict.pop("fname")
    return calc_pds(fname, **argdict)


def calc_pds(lcfile, fftlen,
             save_dyn=False,
             bintime=1,
             pdsrebin=1,
             normalization='leahy',
             back_ctrate=0.,
             noclobber=False,
             outname=None):
    """Calculate the PDS from an input light curve file.

    Parameters
    ----------
    lcfile : str
        The light curve file
    fftlen : float
        The length of the chunks over which FFTs will be calculated, in seconds

    Other Parameters
    ----------------
    save_dyn : bool
        If True, save the dynamical power spectrum
    bintime : float
        The bin time. If different from that of the light curve, a rebinning is
        performed
    pdsrebin : int
        Rebin the PDS of this factor.
    normalization: str
        'Leahy', 'frac', 'rms', or any normalization accepted by ``stingray``.
        Default 'Leahy'
    back_ctrate : float
        The non-source count rate
    noclobber : bool
        If True, do not overwrite existing files
    outname : str
        If speficied, output file name. If not specified or None, the new file
        will have the same root as the input light curve and the '_pds' suffix
    """
    root = hen_root(lcfile)
    outname = root + '_pds' + HEN_FILE_EXTENSION
    if noclobber and os.path.exists(outname):
        print('File exists, and noclobber option used. Skipping')
        return

    logging.info("Loading file %s..." % lcfile)
    lc = load_lcurve(lcfile)
    instr = lc.instr

    if bintime > lc.dt:
        lcrebin = np.rint(bintime / lc.dt)
        logging.info("Rebinning lcs by a factor %d" % lcrebin)
        lc = lc.rebin(lcrebin)
        lc.instr = instr

    pds = AveragedPowerspectrum(lc, segment_size=fftlen,
                                norm=normalization.lower())

    if pdsrebin is not None and pdsrebin != 1:
        pds = pds.rebin(pdsrebin)

    pds.instr = instr
    pds.fftlen = fftlen
    pds.back_phots = back_ctrate * fftlen
    pds.mjdref = lc.mjdref

    logging.info('Saving PDS to %s' % outname)
    save_pds(pds, outname)


def calc_cpds(lcfile1, lcfile2, fftlen,
              save_dyn=False,
              bintime=1,
              pdsrebin=1,
              outname='cpds' + HEN_FILE_EXTENSION,
              normalization='leahy',
              back_ctrate=0.,
              noclobber=False):
    """Calculate the CPDS from a pair of input light curve files.

    Parameters
    ----------
    lcfile1 : str
        The first light curve file
    lcfile2 : str
        The second light curve file
    fftlen : float
        The length of the chunks over which FFTs will be calculated, in seconds

    Other Parameters
    ----------------
    save_dyn : bool
        If True, save the dynamical power spectrum
    bintime : float
        The bin time. If different from that of the light curve, a rebinning is
        performed
    pdsrebin : int
        Rebin the PDS of this factor.
    normalization : str
        'Leahy', 'frac', 'rms', or any normalization accepted by ``stingray``.
        Default 'Leahy'
    back_ctrate : float
        The non-source count rate
    noclobber : bool
        If True, do not overwrite existing files
    outname : str
        Output file name for the cpds. Default: cpds.[nc|p]
    """
    if noclobber and os.path.exists(outname):
        print('File exists, and noclobber option used. Skipping')
        return

    logging.info("Loading file %s..." % lcfile1)
    lc1 = load_lcurve(lcfile1)
    logging.info("Loading file %s..." % lcfile2)
    lc2 = load_lcurve(lcfile2)
    instr1 = lc1.instr
    instr2 = lc2.instr

    assert lc1.dt == lc2.dt, 'Light curves are sampled differently'
    dt = lc1.dt

    if bintime > dt:
        lcrebin = np.rint(bintime / dt)
        logging.info("Rebinning lcs by a factor %d" % lcrebin)
        lc1 = lc1.rebin(lcrebin)
        lc1.instr = instr1
        lc2 = lc2.rebin(lcrebin)
        lc2.instr = instr2

    if lc1.mjdref != lc2.mjdref:
        lc2 = lc2.change_mjdref(lc1.mjdref)

    ctrate = np.sqrt(lc1.meanrate * lc2.meanrate)

    cpds = AveragedCrossspectrum(lc1, lc2, segment_size=fftlen,
                                 norm=normalization.lower())

    if pdsrebin is not None and pdsrebin != 1:
        cpds = cpds.rebin(pdsrebin)

    cpds.instrs = instr1 + ',' + instr2
    cpds.fftlen = fftlen
    cpds.back_phots = back_ctrate * fftlen
    cpds.mjdref = lc1.mjdref
    lags, lags_err = cpds.time_lag()
    cpds.lag = lags
    cpds.lag_err = lags

    logging.info('Saving CPDS to %s' % outname)
    save_pds(cpds, outname)


def calc_fspec(files, fftlen,
               do_calc_pds=True,
               do_calc_cpds=True,
               do_calc_cospectrum=True,
               do_calc_lags=True,
               save_dyn=False,
               bintime=1,
               pdsrebin=1,
               outroot=None,
               normalization='leahy',
               nproc=1,
               back_ctrate=0.,
               noclobber=False,
               ignore_instr=False):
    r"""Calculate the frequency spectra: the PDS, the cospectrum, ...

    Parameters
    ----------
    files : list of str
        List of input file names
    fftlen : float
        length of chunks to perform the FFT on.

    Other Parameters
    ----------------
    save_dyn : bool
        If True, save the dynamical power spectrum
    bintime : float
        The bin time. If different from that of the light curve, a rebinning is
        performed
    pdsrebin : int
        Rebin the PDS of this factor.
    normalization : str
        'Leahy' [3] or 'rms' [4] [5]. Default 'Leahy'.
    back_ctrate : float
        The non-source count rate
    noclobber : bool
        If True, do not overwrite existing files
    outroot : str
        Output file name root
    nproc : int
        Number of processors to use to parallelize the processing of multiple
        files
    ignore_instr : bool
        Ignore instruments; files are alternated in the two channels

    References
    ----------
    [3] Leahy et al. 1983, ApJ, 266, 160.

    [4] Belloni & Hasinger 1990, A&A, 230, 103

    [5] Miyamoto et al. 1991, ApJ, 383, 784

    """

    logging.info('Using %s normalization' % normalization)

    if do_calc_pds:
        wrapped_file_dicts = []
        for f in files:
            wfd = {"fftlen": fftlen,
                   "save_dyn": save_dyn,
                   "bintime": bintime,
                   "pdsrebin": pdsrebin,
                   "normalization": normalization.lower(),
                   "back_ctrate": back_ctrate,
                   "noclobber": noclobber}
            wfd["fname"] = f
            wrapped_file_dicts.append(wfd)

        if os.name == 'nt' or nproc == 1:
            [_wrap_fun_pds(w) for w in wrapped_file_dicts]
        else:
            pool = Pool(processes=nproc)
            for i in pool.imap_unordered(_wrap_fun_pds, wrapped_file_dicts):
                pass
            pool.close()

    if not do_calc_cpds or len(files) < 2:
        return

    if ignore_instr:
        files1 = files[0::2]
        files2 = files[1::2]
    else:
        logging.info('Sorting file list')
        sorted_files = sort_files(files)

        logging.warning('Beware! For cpds and derivatives, I assume that the '
                        'files are from only two instruments and in pairs '
                        '(even in random order)')

        instrs = list(sorted_files.keys())

        files1 = sorted_files[instrs[0]]
        files2 = sorted_files[instrs[1]]

    assert len(files1) == len(files2), 'An even number of files is needed'

    argdict = {"fftlen": fftlen,
               "save_dyn": save_dyn,
               "bintime": bintime,
               "pdsrebin": pdsrebin,
               "normalization": normalization.lower(),
               "back_ctrate": back_ctrate,
               "noclobber": noclobber}

    funcargs = []

    for i_f, f in enumerate(files1):
        f1, f2 = f, files2[i_f]

        outdir = os.path.dirname(f1)
        if outdir == '':
            outdir = os.getcwd()

        outr = _assign_value_if_none(
            outroot,
            common_name(f1, f2, default='%d' % i_f))

        outname = os.path.join(outdir,
                               outr.replace(HEN_FILE_EXTENSION, '') +
                               '_cpds' + HEN_FILE_EXTENSION)

        funcargs.append([f1, f2, outname, argdict])

    if os.name == 'nt' or nproc == 1:
        [_wrap_fun_cpds(fa) for fa in funcargs]
    else:
        pool = Pool(processes=nproc)
        for i in pool.imap_unordered(_wrap_fun_cpds, funcargs):
            pass
        pool.close()


def _normalize(array, ref=0):
    m = ref
    std = np.std(array)
    newarr = np.zeros_like(array)
    good = array > m
    newarr[good] = (array[good] - ref) / std
    return newarr


def dumpdyn(fname, plot=False):
    """Dump a dynamical frequency spectrum in text files.

    Parameters
    ----------
    fname : str
        The file name

    Other Parameters
    ----------------
    plot : bool
        if True, plot the spectrum

    """
    ftype, pdsdata = get_file_type(fname, specify_reb=False)

    dynpds = pdsdata['dyn' + ftype]
    edynpds = pdsdata['edyn' + ftype]

    try:
        freq = pdsdata['freq']
    except:
        flo = pdsdata['flo']
        fhi = pdsdata['fhi']
        freq = (fhi + flo) / 2

    time = pdsdata['dyntimes']
    freqs = np.zeros_like(dynpds)
    times = np.zeros_like(dynpds)

    for i, im in enumerate(dynpds):
        freqs[i, :] = freq
        times[i, :] = time[i]

    t = times.real.flatten()
    f = freqs.real.flatten()
    d = dynpds.real.flatten()
    e = edynpds.real.flatten()

    np.savetxt('{0}_dumped_{1}.txt'.format(hen_root(fname), ftype),
               np.array([t, f, d, e]).T)
    size = _normalize(d)
    if plot:
        import matplotlib.pyplot as plt
        plt.scatter(t, f, s=size)
        plt.xlabel('Time (s)')
        plt.ylabel('Freq (Hz)')

        plt.show()


def dumpdyn_main(args=None):
    """Main function called by the `HENdumpdyn` command line script."""
    import argparse

    description = ('Dump dynamical (cross) power spectra')
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help=("List of files in any valid HENDRICS "
                                       "format for PDS or CPDS"), nargs='+')
    parser.add_argument("--noplot", help="plot results",
                        default=False, action='store_true')

    args = parser.parse_args(args)

    fnames = args.files

    for f in fnames:
        dumpdyn(f, plot=not args.noplot)


def main(args=None):
    """Main function called by the `HENfspec` command line script."""
    import argparse
    description = ('Create frequency spectra (PDS, CPDS, cospectrum) '
                   'starting from well-defined input ligthcurves')
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of light curve files", nargs='+')
    parser.add_argument("-b", "--bintime", type=float, default=1/4096,
                        help="Light curve bin time; if negative, interpreted" +
                        " as negative power of 2." +
                        " Default: 2^-10, or keep input lc bin time" +
                        " (whatever is larger)")
    parser.add_argument("-r", "--rebin", type=int, default=1,
                        help="(C)PDS rebinning to apply. Default: none")
    parser.add_argument("-f", "--fftlen", type=float, default=512,
                        help="Length of FFTs. Default: 512 s")
    parser.add_argument("-k", "--kind", type=str, default="PDS,CPDS,cos",
                        help='Spectra to calculate, as comma-separated list' +
                        ' (Accepted: PDS and CPDS;' +
                        ' Default: "PDS,CPDS")')
    parser.add_argument("--norm", type=str, default="leahy",
                        help='Normalization to use' +
                        ' (Accepted: leahy and rms;' +
                        ' Default: "leahy")')
    parser.add_argument("--noclobber", help="Do not overwrite existing files",
                        default=False, action='store_true')
    parser.add_argument("-o", "--outroot", type=str, default=None,
                        help='Root of output file names for CPDS only')
    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING',
                        type=str)
    parser.add_argument("--nproc",
                        help=("Number of processors to use"),
                        default=1,
                        type=int)
    parser.add_argument("--back",
                        help=("Estimated background (non-source) count rate"),
                        default=0.,
                        type=float)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')
    parser.add_argument("--save-dyn", help="save dynamical power spectrum",
                        default=False, action='store_true')
    parser.add_argument("--ignore-instr",
                        help="Ignore instrument names in channels",
                        default=False, action='store_true')

    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = 'DEBUG'

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENfspec.log', level=numeric_level,
                        filemode='w')

    bintime = args.bintime
    fftlen = args.fftlen
    pdsrebin = args.rebin
    normalization = args.norm
    if normalization.lower() not in ["frac", "abs", "leahy", "none", "rms"]:
        warnings.warn('Beware! Unknown normalization!')
        normalization = 'leahy'
    if normalization == 'rms':
        normalization = 'frac'

    do_cpds = do_pds = do_cos = do_lag = False
    kinds = args.kind.split(',')
    for k in kinds:
        if k == 'PDS':
            do_pds = True
        elif k == 'CPDS':
            do_cpds = True
        elif k == 'cos' or k == 'cospectrum':
            do_cos = True
            do_cpds = True
        elif k == 'lag':
            do_lag = True
            do_cpds = True

    calc_fspec(args.files, fftlen,
               do_calc_pds=do_pds,
               do_calc_cpds=do_cpds,
               do_calc_cospectrum=do_cos,
               do_calc_lags=do_lag,
               save_dyn=args.save_dyn,
               bintime=bintime,
               pdsrebin=pdsrebin,
               outroot=args.outroot,
               normalization=normalization,
               nproc=args.nproc,
               back_ctrate=args.back,
               noclobber=args.noclobber,
               ignore_instr=args.ignore_instr)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""This is proposed as an Astropy affiliated package."""

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *
# ----------------------------------------------------------------------------

# For egg_info test builds to pass, put package imports here.
if not _ASTROPY_SETUP_:
    from . import base
    from . import binary
    from . import calibrate
    from . import colors
    from . import create_gti
    from . import efsearch
    from . import exposure
    from . import exvar
    from . import fake
    from . import fspec
    from . import io
    from . import lcurve
    from . import modeling
    from . import plot
    from . import read_events
    from . import rebin
    from . import save_as_xspec
    from . import sum_fspec
    from . import timelags
    from . import varenergy



#!/usr/bin/env python
from __future__ import (division, print_function, absolute_import)

import argparse

import numpy as np
import matplotlib.pyplot as plt
import astropy.io.fits as pf

import warnings
from .io import is_string
from stingray.io import load_events_and_gtis, ref_mjd
from .base import _assign_value_if_none, hen_root
from .fold import fit_profile, std_fold_fit_func

from stingray.pulse.pulsar import pulse_phase, phase_exposure


def outfile_name(file):
    return file.replace(".evt", "_phasetag.evt")


def phase_tag(ev_list, parameter_info, gtis=None, mjdref=0,
              nbin=10, ref_to_max=False, pepoch=None,
              expocorr=True, pulse_ref_time=None, plot=True, test=False):
    """Phase-tag events in a FITS file with a given ephemeris.

    Parameters
    ----------
    ev_list : float
        Event times
    parameter_info : str or array of floats
        If a string, this is a pulsar parameter file that PINT is able to
        understand. Otherwise, this is a list of frequency derivatives
        [F0, F1, F2, ...]

    Other parameters
    ----------------
    gtis : [[g0_0, g0_1], [g1_0, g1_1], ...]
        Good time intervals
    nbin : int
        Number of nbin in the pulsed profile
    ref_to_max : bool
        Automatically refer the TOAs to the maximum of the profile
    pepoch : float, default None
        Reference epoch for the timing solution. If None, this is the start
        of the observation.
    pulse_ref_time : float
        Reference time for the pulse. This overrides ref_to_max
    plot : bool
        Plot diagnostics
    expocorr : bool
        Use exposure correction when calculating the profile

    """
    # ---- in MJD ----
    if gtis is None:
        gtis = np.array([[ev_list[0], ev_list[-1]]])

    ev_mjd = ev_list / 86400 + mjdref
    gtis_mjd = gtis / 86400 + mjdref

    pepoch = _assign_value_if_none(pepoch, gtis_mjd[0, 0])

    # ------ Orbital DEMODULATION --------------------
    if is_string(parameter_info):
        raise NotImplementedError('This part is not yet implemented. Please '
                                  'use single frequencies and pepoch as '
                                  'documented')

    else:
        frequency_derivatives = parameter_info
        times = (ev_mjd - pepoch) * 86400

    f = frequency_derivatives[0]

    phase = pulse_phase(times, *frequency_derivatives, to_1=False)
    gti_phases = pulse_phase((gtis_mjd - pepoch) * 86400,
                             *frequency_derivatives, to_1=False)

    # ------- now apply period derivatives ------

    print("Calculating phases...", end='')
    ref_phase = 0
    ref_time = 0

    if pulse_ref_time is not None:
        ref_time = (pulse_ref_time - pepoch) * 86400
        ref_phase = ref_time * f
    elif ref_to_max:
        phase_to1 = phase - np.floor(phase)

        raw_profile, bins = np.histogram(phase_to1,
                                         bins=np.linspace(0, 1, nbin + 1))
        exposure = phase_exposure(gti_phases[0, 0], gti_phases[-1, 1], 1,
                                  nbin=nbin, gtis=gti_phases)
        profile = raw_profile / exposure
        profile_err = np.sqrt(raw_profile) / exposure

        sinpars, bu, bu = fit_profile(profile, profile_err,
                                      nperiods=2, baseline=True,
                                      debug=test)
        fine_phases = np.linspace(0, 2, 1000 * 2)
        fitted_profile = std_fold_fit_func(sinpars, fine_phases)
        maxp = np.argmax(fitted_profile)
        ref_phase = fine_phases[maxp]
        if test:  # pragma: no cover
            # No tests with a pulsed profile yet
            ref_phase = bins[np.argmax(raw_profile)]

        ref_time = ref_phase / f

    phase -= ref_phase
    gti_phases -= ref_phase
    phase_to1 = phase - np.floor(phase)

    raw_profile, bins = np.histogram(phase_to1,
                                     bins=np.linspace(0, 1, nbin + 1))

    exposure = phase_exposure(gti_phases[0, 0], gti_phases[-1, 1], 1,
                              nbin=nbin, gtis=gti_phases)
    if np.any(np.logical_or(exposure != exposure, exposure == 0)):
        warnings.warn('Exposure has NaNs or zeros. Profile is not normalized')
        expocorr = False

    if not expocorr:
        exposure = np.ones_like(raw_profile)

    profile = raw_profile / exposure
    profile = np.append(profile, profile)
    exposure = np.append(exposure, exposure)
    profile_err = np.sqrt(profile)
    phs = (bins[1:] + bins[:-1]) / 2
    phs = np.append(phs, phs + 1)

    fig = None
    if plot:
        fig = plt.figure()
        plt.errorbar(phs, profile / exposure,
                     yerr=profile_err / exposure, fmt='none')
        plt.plot(phs, profile / exposure, 'k-',
                 drawstyle='steps-mid')
        plt.xlabel("Phase")
        plt.ylabel("Counts")
        for i in range(20):
            plt.axvline(i * 0.1, ls='--', color='b')
        if not test:  # pragma: no cover
            plt.show()
    # ------ WRITE RESULTS BACK TO FITS --------------
    results = type('results', (object,), {})
    results.ev_list = ev_list
    results.phase = phase
    results.frequency_derivatives = frequency_derivatives
    results.ref_time = ref_time
    results.figure = fig
    return results


def phase_tag_fits(filename, parameter_info, **kwargs):
    """Phase-tag events in a FITS file with a given ephemeris.

    Parameters
    ----------
    filename : str
        Events FITS file
    parameter_info : str or array of floats
        If a string, this is a pulsar parameter file that PINT is able to
        understand. Otherwise, this is a list of frequency derivatives
        [F0, F1, F2, ...]

    Other parameters
    ----------------
    nbin : int
        Number of nbin in the pulsed profile
    ref_to_max : bool
        Automatically refer the TOAs to the maximum of the profile
    pepoch : float, default None
        Reference epoch for the timing solution. If None, this is the start
        of the observation.
    pulse_ref_time : float
        Reference time for the pulse. This overrides ref_to_max
    plot : bool
        Plot diagnostics
    expocorr : bool
        Use exposure correction when calculating the profile

    """

    outfile = outfile_name(filename)
    evreturns = load_events_and_gtis(filename)
    mjdref = ref_mjd(filename)

    results = phase_tag(evreturns.ev_list, parameter_info,
                        gtis=evreturns.gti_list,
                        mjdref=mjdref,
                        **kwargs)
    if results.figure is not None:
        results.figure.savefig(hen_root(filename) + '.pdf')
    phase = results.phase
    frequency_derivatives = results.frequency_derivatives
    ref_time = results.ref_time

    phase_to1 = phase - np.floor(phase)

    # Save results to fits file
    hdulist = pf.open(filename, checksum=True)
    tbhdu = hdulist["EVENTS"]
    table = tbhdu.data
    order = np.argsort(table['TIME'])
    # load_events_and_gtis sorts the data automatically. This is not the case
    # of the other operations done here. So, let's sort the table first.
    for col in table.names:
        table.field(col)[:] = np.array(table[col])[order]

    # If columns already exist, overwrite them. Else, create them
    create = False
    if 'Orbit_bary' in table.names:
        table.field("Orbit_bary")[:] = evreturns.ev_list
        table.field("TotPhase_s")[:] = \
            phase / frequency_derivatives[0] + ref_time
        table.field("Phase")[:] = phase_to1
    else:
        create = True

    # first of all, copy columns
    cols = table.columns

    # create new list of columns, copying all columns from other table
    newlist = []

    for c in cols:
        newlist.append(c)

    if create:
        # then, create new column with orbital demodulation
        newcol = pf.Column(name='Orbit_bary',
                           format='1D',
                           unit='s',
                           array=results.ev_list)

        # append it to new table
        newlist.append(newcol)

        # Do the same with total phase
        newcol = pf.Column(name='TotPhase_s',
                           format='1D',
                           unit='s',
                           array=phase / frequency_derivatives[0] + ref_time)

        newlist.append(newcol)

        # Do the same with fractional phase
        newcol = pf.Column(name='Phase',
                           format='1D',
                           unit='phase',
                           array=phase_to1)

        newlist.append(newcol)

        # make new column definitions
    coldefs = pf.ColDefs(newlist)

    # create new record
    newrec = pf.FITS_rec.from_columns(coldefs)

    # and new hdu
    newtbhdu = pf.BinTableHDU(data=newrec,
                              header=tbhdu.header.copy(),
                              name='EVENTS', uint=False)

    # Copy primary HDU from old file
    prihdu = hdulist[0].copy()

    # define new hdulist
    newhdulist = pf.HDUList([prihdu, newtbhdu])

    # Copy remaining HDUs from old file
    if len(hdulist) > 2:
        for h in hdulist[2:]:
            newhdulist.append(h.copy())

    newhdulist.writeto(outfile, overwrite=True)
    hdulist.close()


def main_phasetag(args=None):

    parser = argparse.ArgumentParser()

    parser.add_argument("file", help="Event file", type=str)
    parser.add_argument("--parfile", help="Parameter file", type=str,
                        default=None)
    parser.add_argument('-f', "--freqs", help="Frequency derivatives",
                        type=float, default=None, nargs='+')
    parser.add_argument("-n", "--nbin", type=int, default=16, help="Nbin")
    parser.add_argument("--plot", action="store_true", default=False,
                        dest="plot", help="Plot profile")
    parser.add_argument("--tomax", action="store_true", default=False,
                        dest="tomax",
                        help="Refer phase to pulse max")
    parser.add_argument("--test", action="store_true", default=False,
                        dest="test",
                        help="Only for unit tests! Do not use")
    parser.add_argument("--refTOA", default=None, type=np.longdouble,
                        help="Reference TOA in MJD (overrides --tomax) for "
                             "reference pulse phase",
                        dest='pulse_ref_time')
    parser.add_argument("--pepoch", default=None, type=np.longdouble,
                        help="Reference time for timing solution",
                        dest='pepoch')

    args = parser.parse_args(args)

    if args.freqs is None and args.parfile is None:
        raise ValueError('Specify one between --freqs or --parfile')
    elif args.freqs is not None and args.parfile is not None:
        raise ValueError('Specify only one between --freqs or --parfile')
    elif args.freqs is not None:
        parameter_info = args.freqs
    else:
        parameter_info = args.parfile

    plot = args.plot
    expocorr = True

    phase_tag_fits(args.file, parameter_info, plot=plot, nbin=args.nbin,
                   test=args.test, expocorr=expocorr, ref_to_max=args.tomax,
                   pulse_ref_time=args.pulse_ref_time, pepoch=args.pepoch)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import division, print_function
from stingray.events import EventList
from stingray.lightcurve import Lightcurve
from stingray.powerspectrum import Powerspectrum, AveragedPowerspectrum
from stingray.powerspectrum import Crossspectrum, AveragedCrossspectrum
import numpy as np
import os
from hendrics.io import load_events, save_events, save_lcurve, load_lcurve
from hendrics.io import save_data, load_data, save_pds, load_pds
from hendrics.io import HEN_FILE_EXTENSION, _split_high_precision_number
from hendrics.io import save_model, load_model, HAS_C256, HAS_NETCDF

import pytest
import glob
from astropy.modeling import models
from astropy.modeling.core import Model

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


def _dummy_bad(x, z, y=0):
    return


def _dummy(x, y=0):
    return


class TestIO():
    """Real unit tests."""
    @classmethod
    def setup_class(cls):
        cls.dum = 'bubu' + HEN_FILE_EXTENSION

    def test_save_data(self):
        struct = {'a': 0.1, 'b': np.longdouble('123.4567890123456789'),
                  'c': np.longdouble([[-0.5, 3.5]]),
                  'd': 1}
        save_data(struct, self.dum)
        struct2 = load_data(self.dum)
        assert np.allclose(struct['a'], struct2['a'])
        assert np.allclose(struct['b'], struct2['b'])
        assert np.allclose(struct['c'], struct2['c'])
        assert np.allclose(struct['d'], struct2['d'])

    def test_load_and_save_events(self):
        events = EventList([0, 2, 3.], pi=[1, 2, 3], mjdref=54385.3254923845,
                           gti=np.longdouble([[-0.5, 3.5]]))
        events.energy = np.array([3., 4., 5.])
        save_events(events, self.dum)
        events2 = load_events(self.dum)
        assert np.allclose(events.time, events2.time)
        assert np.allclose(events.pi, events2.pi)
        assert np.allclose(events.mjdref, events2.mjdref)
        assert np.allclose(events.gti, events2.gti)
        assert np.allclose(events.energy, events2.energy)

    def test_load_and_save_lcurve(self):
        lcurve = Lightcurve(np.linspace(0, 10, 15), np.random.poisson(30, 15),
                            mjdref=54385.3254923845,
                            gti=np.longdouble([[-0.5, 3.5]]))
        save_lcurve(lcurve, self.dum)
        lcurve2 = load_lcurve(self.dum)
        assert np.allclose(lcurve.time, lcurve2.time)
        assert np.allclose(lcurve.counts, lcurve2.counts)
        assert np.allclose(lcurve.mjdref, lcurve2.mjdref)
        assert np.allclose(lcurve.gti, lcurve2.gti)
        assert lcurve.err_dist == lcurve2.err_dist

    def test_load_and_save_pds(self):
        pds = Powerspectrum()
        pds.freq = np.linspace(0, 10, 15)
        pds.power = np.random.poisson(30, 15)
        pds.mjdref = 54385.3254923845
        pds.gti = np.longdouble([[-0.5, 3.5]])

        save_pds(pds, self.dum)
        pds2 = load_pds(self.dum)
        assert np.allclose(pds.gti, pds2.gti)
        assert np.allclose(pds.mjdref, pds2.mjdref)
        assert pds.m == pds2.m

    def test_load_and_save_xps(self):
        lcurve1 = Lightcurve(np.linspace(0, 10, 150),
                             np.random.poisson(30, 150),
                             mjdref=54385.3254923845,
                             gti=np.longdouble([[-0.5, 3.5]]))
        lcurve2 = Lightcurve(np.linspace(0, 10, 150),
                             np.random.poisson(30, 150),
                             mjdref=54385.3254923845,
                             gti=np.longdouble([[-0.5, 3.5]]))

        xps = AveragedCrossspectrum(lcurve1, lcurve2, 1)

        save_pds(xps, self.dum)
        xps2 = load_pds(self.dum)
        assert np.allclose(xps.gti, xps2.gti)
        assert xps.m == xps2.m
        lag, lag_err = xps.time_lag()
        lag2, lag2_err = xps2.time_lag()
        assert np.allclose(lag, lag2)
        assert hasattr(xps2, 'pds1')

    def test_load_and_save_xps_no_all(self):
        lcurve1 = Lightcurve(np.linspace(0, 10, 150),
                             np.random.poisson(30, 150),
                             mjdref=54385.3254923845,
                             gti=np.longdouble([[-0.5, 3.5]]))
        lcurve2 = Lightcurve(np.linspace(0, 10, 150),
                             np.random.poisson(30, 150),
                             mjdref=54385.3254923845,
                             gti=np.longdouble([[-0.5, 3.5]]))

        xps = AveragedCrossspectrum(lcurve1, lcurve2, 1)

        outfile = 'small_xps' + HEN_FILE_EXTENSION
        save_pds(xps, outfile, save_all=False)
        xps2 = load_pds(outfile)
        assert np.allclose(xps.gti, xps2.gti)
        assert xps.m == xps2.m
        assert not hasattr(xps2, 'pds1')

    def test_load_and_save_xps_quick(self):
        lcurve1 = Lightcurve(np.linspace(0, 10, 150),
                             np.random.poisson(30, 150),
                             mjdref=54385.3254923845,
                             gti=np.longdouble([[-0.5, 3.5]]))
        lcurve2 = Lightcurve(np.linspace(0, 10, 150),
                             np.random.poisson(30, 150),
                             mjdref=54385.3254923845,
                             gti=np.longdouble([[-0.5, 3.5]]))

        xps = AveragedCrossspectrum(lcurve1, lcurve2, 1)

        save_pds(xps, self.dum)
        xps2 = load_pds(self.dum, nosub=True)
        assert np.allclose(xps.gti, xps2.gti)
        assert xps.m == xps2.m

        assert not hasattr(xps2, 'pds1')

    def test_high_precision_split1(self):
        C_I, C_F, C_l, k = \
            _split_high_precision_number("C", np.double(0.01), 8)
        assert C_I == 1
        np.testing.assert_almost_equal(C_F, 0, 6)
        assert C_l == -2
        assert k == "double"

    def test_high_precision_split2(self):
        C_I, C_F, C_l, k = \
            _split_high_precision_number("C", np.double(1.01), 8)
        assert C_I == 1
        np.testing.assert_almost_equal(C_F, np.double(0.01), 6)
        assert C_l == 0
        assert k == "double"

    def test_save_longcomplex(self):
        val = np.longcomplex(1.01 + 2.3j)
        data = {'val': val}
        save_data(data, 'bubu' + HEN_FILE_EXTENSION)
        data_out = load_data('bubu' + HEN_FILE_EXTENSION)

        assert np.allclose(data['val'], data_out['val'])

    @pytest.mark.skipif('not HAS_C256 or not HAS_NETCDF')
    def test_save_longcomplex(self):
        val = np.complex256(1.01 + 2.3j)
        data = {'val': val}
        with pytest.warns(UserWarning) as record:
            save_data(data, 'bubu' + HEN_FILE_EXTENSION)
        assert "complex256 yet unsupported" in record[0].message.args[0]
        data_out = load_data('bubu' + HEN_FILE_EXTENSION)

        assert np.allclose(data['val'], data_out['val'])

    @classmethod
    def teardown_class(cls):
        import shutil
        for dum in glob.glob('bubu*.*'):
            os.unlink(dum)
        shutil.rmtree('bubu')


class TestIOModel():
    """Real unit tests."""

    @classmethod
    def setup_class(cls):
        cls.dum = 'bubu' + HEN_FILE_EXTENSION
        cls.model = models.Gaussian1D() + models.Const1D(amplitude=2)

    def test_save_and_load_Astropy_model(self):
        save_model(self.model, 'bubu_model.p')
        b, kind, _ = load_model('bubu_model.p')
        assert kind == 'Astropy'
        assert isinstance(b, Model)
        assert np.all(self.model.parameters == b.parameters)
        assert np.all(self.model.bounds == b.bounds)
        assert np.all(self.model.fixed == b.fixed)

    def test_save_and_load_callable_model(self):
        constraints0 = {'bounds': ()}
        save_model(_dummy, 'bubu_callable.p', constraints=constraints0)
        b, kind, constraints = load_model('bubu_callable.p')
        assert kind == 'callable'
        assert callable(b)
        assert np.all(_dummy.__code__.co_argcount == b.__code__.co_argcount)
        assert np.all(_dummy.__defaults__ == b.__defaults__)
        assert np.all(constraints == constraints0)

    def test_save_callable_model_wrong(self):
        with pytest.raises(TypeError) as record:
            save_model(_dummy_bad, 'callable_bad.p')
        assert 'Accepted callable models have only' in str(record.value)
        assert not os.path.exists('callable_bad.p')

    def test_save_junk_model(self):
        a = 'g'
        with pytest.raises(TypeError) as record:
            save_model(a, 'bad.p', constraints={'bounds': ()})
        assert 'The model has to be an Astropy model or a callable' \
               in str(record.value)
        assert not os.path.exists('bad.p')

    def test_load_python_model_callable(self):
        modelstring = '''
def model(x, a=2, b=4):
    return x * a + b

constraints = {'fixed': {'a': True}}
'''
        with open('bubu__model__.py', 'w') as fobj:
            print(modelstring, file=fobj)
        b, kind, constraints = load_model('bubu__model__.py')
        assert kind == 'callable'
        assert callable(b)
        assert b.__code__.co_argcount == 3
        assert b.__defaults__[0] == 2
        assert b.__defaults__[1] == 4
        assert np.all(constraints == {'fixed': {'a': True}})

    def test_load_python_model_Astropy(self):
        modelstring = '''
from astropy.modeling import models
model = models.Const1D()
'''
        with open('bubu__model__2__.py', 'w') as fobj:
            print(modelstring, file=fobj)
        b, kind, constraints = load_model('bubu__model__2__.py')
        assert isinstance(b, Model)
        assert kind == 'Astropy'
        assert b.amplitude == 1

    def test_load_model_input_not_string(self):
        '''Input is not a string'''
        with pytest.raises(TypeError) as record:
            b, kind, _ = load_model(1)
        assert 'modelstring has to be an existing file name' \
               in str(record.value)

    def test_load_model_input_file_doesnt_exist(self):
        with pytest.raises(FileNotFoundError) as record:
            b, kind, _ = load_model('dfasjkdaslfj')
        assert 'Model file not found' in str(record.value)

    def test_load_model_input_invalid_file_format(self):
        with open('bubu.txt', 'w') as fobj:
            print(1, file=fobj)
        with pytest.raises(TypeError) as record:
            b, kind, _ = load_model('bubu.txt')
        assert 'Unknown file type' in str(record.value)

    def test_load_data_fails(self):
        with pytest.raises(TypeError) as record:
            load_data('afile.fits')
        assert "The file type is not recognized" in str(record.value)

    @classmethod
    def teardown_class(cls):
        import shutil
        for dum in glob.glob('bubu*.*'):
            os.unlink(dum)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""A miscellaneous collection of basic functions."""

from __future__ import (absolute_import, unicode_literals, division,
                        print_function)

import numpy as np
import logging
import sys


def r_in(td, r_0):
    """Calculate incident countrate given dead time and detected countrate."""
    tau = 1 / r_0
    return 1. / (tau - td)


def r_det(td, r_in):
    """Calculate detected countrate given dead time and incident countrate."""
    tau = 1 / r_in
    return 1. / (tau + td)


def _assign_value_if_none(value, default):
    if value is None:
        return default
    else:
        return value


def _look_for_array_in_array(array1, array2):
    for a1 in array1:
        if a1 in array2:
            return a1


def is_string(s):
    """Portable function to answer this question."""
    PY3 = sys.version_info[0] == 3
    if PY3:
        return isinstance(s, str)  # NOQA
    else:
        return isinstance(s, basestring)  # NOQA


def _order_list_of_arrays(data, order):
    if hasattr(data, 'items'):
        data = dict([(i[0], i[1][order])
                     for i in data.items()])
    elif hasattr(data, 'index'):
        data = [i[order] for i in data]
    else:
        data = None
    return data


class _empty():
    def __init__(self):
        pass


def mkdir_p(path):  # pragma: no cover
    """Safe mkdir function.

    Parameters
    ----------
    path : str
        Name of the directory/ies to create

    Notes
    -----
    Found at
    http://stackoverflow.com/questions/600268/mkdir-p-functionality-in-python
    """
    import os
    import errno
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def read_header_key(fits_file, key, hdu=1):
    """Read the header key key from HDU hdu of the file fits_file.

    Parameters
    ----------
    fits_file: str
    key: str
        The keyword to be read

    Other Parameters
    ----------------
    hdu : int
    """
    from astropy.io import fits as pf

    hdulist = pf.open(fits_file)
    try:
        value = hdulist[hdu].header[key]
    except:  # pragma: no cover
        value = ''
    hdulist.close()
    return value


def ref_mjd(fits_file, hdu=1):
    """Read MJDREFF+ MJDREFI or, if failed, MJDREF, from the FITS header.

    Parameters
    ----------
    fits_file : str

    Returns
    -------
    mjdref : numpy.longdouble
        the reference MJD

    Other Parameters
    ----------------
    hdu : int
    """
    import collections

    if isinstance(fits_file, collections.Iterable) and\
            not is_string(fits_file):
        fits_file = fits_file[0]
        logging.info("opening %s" % fits_file)

    try:
        ref_mjd_int = np.long(read_header_key(fits_file, 'MJDREFI'))
        ref_mjd_float = np.longdouble(read_header_key(fits_file, 'MJDREFF'))
        ref_mjd_val = ref_mjd_int + ref_mjd_float
    except:
        ref_mjd_val = np.longdouble(read_header_key(fits_file, 'MJDREF'))
    return ref_mjd_val


def common_name(str1, str2, default='common'):
    """Strip two strings of the letters not in common.

    Filenames must be of same length and only differ by a few letters.

    Parameters
    ----------
    str1 : str
    str2 : str

    Returns
    -------
    common_str : str
        A string containing the parts of the two names in common

    Other Parameters
    ----------------
    default : str
        The string to return if common_str is empty
    """
    if not len(str1) == len(str2):
        return default
    common_str = ''
    # Extract the HEN root of the name (in case they're event files)
    str1 = hen_root(str1)
    str2 = hen_root(str2)
    for i, letter in enumerate(str1):
        if str2[i] == letter:
            common_str += letter
    # Remove leading and trailing underscores and dashes
    common_str = common_str.rstrip('_').rstrip('-')
    common_str = common_str.lstrip('_').lstrip('-')
    if common_str == '':
        common_str = default
    logging.debug('common_name: %s %s -> %s' % (str1, str2, common_str))
    return common_str


def hen_root(filename):
    """Return the root file name (without _ev, _lc, etc.).

    Parameters
    ----------
    filename : str
    """
    import os.path
    fname = filename.replace('.gz', '')
    fname = os.path.splitext(fname)[0]
    fname = fname.replace('_ev', '').replace('_lc', '')
    fname = fname.replace('_calib', '')
    return fname


def optimal_bin_time(fftlen, tbin):
    """Vary slightly the bin time to have a power of two number of bins.

    Given an FFT length and a proposed bin time, return a bin time
    slightly shorter than the original, that will produce a power-of-two number
    of FFT bins.
    """
    import numpy as np
    return fftlen / (2 ** np.ceil(np.log2(fftlen / tbin)))


def detection_level(nbins, epsilon=0.01, n_summed_spectra=1, n_rebin=1):
    r"""Detection level for a PDS.

    Return the detection level (with probability 1 - epsilon) for a Power
    Density Spectrum of nbins bins, normalized a la Leahy (1983), based on
    the 2-dof :math:`{\chi}^2` statistics, corrected for rebinning (n_rebin)
    and multiple PDS averaging (n_summed_spectra)
    """
    try:
        from scipy import stats
    except:  # pragma: no cover
        raise Exception('You need Scipy to use this function')

    import collections
    if not isinstance(n_rebin, collections.Iterable):
        r = n_rebin
        retlev = stats.chi2.isf(epsilon / nbins, 2 * n_summed_spectra * r) \
            / (n_summed_spectra * r)
    else:
        retlev = [stats.chi2.isf(epsilon / nbins, 2 * n_summed_spectra * r) /
                  (n_summed_spectra * r) for r in n_rebin]
        retlev = np.array(retlev)
    return retlev


def probability_of_power(level, nbins, n_summed_spectra=1, n_rebin=1):
    r"""Give the probability of a given power level in PDS.

    Return the probability of a certain power level in a Power Density
    Spectrum of nbins bins, normalized a la Leahy (1983), based on
    the 2-dof :math:`{\chi}^2` statistics, corrected for rebinning (n_rebin)
    and multiple PDS averaging (n_summed_spectra)
    """
    try:
        from scipy import stats
    except:  # pragma: no cover
        raise Exception('You need Scipy to use this function')

    epsilon = nbins * stats.chi2.sf(level * n_summed_spectra * n_rebin,
                                    2 * n_summed_spectra * n_rebin)
    return 1 - epsilon


def gti_len(gti):
    """Return the total good time from a list of GTIs."""
    return np.sum([g[1] - g[0] for g in gti])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions to create and apply GTIs."""

from __future__ import (absolute_import, unicode_literals, division,
                        print_function)

from .io import HEN_FILE_EXTENSION, save_data, load_data, get_file_type
from stingray.gti import cross_gtis, create_gti_from_condition, create_gti_mask
from .base import hen_root, _assign_value_if_none
import logging
import numpy as np
import sys


def filter_gti_by_length(gti, minimum_length):
    """Filter a list of GTIs: keep those longer than `minimum_length`."""
    if minimum_length == 0 or minimum_length is None:
        return gti

    newgtis = []
    for g in gti:
        length = g[1] - g[0]
        if length >= minimum_length:
            newgtis.append(g)

    return np.array(newgtis)


def create_gti(fname, filter_expr, safe_interval=[0, 0], outfile=None,
               minimum_length=0):
    """Create a GTI list by using boolean operations on file data.

    Parameters
    ----------
    fname : str
        File name. The file must be in HENDRICS format.
    filter_expr : str
        A boolean condition on one or more of the arrays contained in the data.
        E.g. '(lc > 10) & (lc < 20)'

    Returns
    -------
    gtis : [[gti0_0, gti0_1], [gti1_0, gti1_1], ...]
        The newly created GTIs

    Other parameters
    ----------------
    safe_interval : float or [float, float]
        A safe interval to exclude at both ends (if single float) or the start
        and the end (if pair of values) of GTIs.
    outfile : str
        The output file name. If None, use a default root + '_gti' combination
    """
    # Necessary as nc variables are sometimes defined as array
    from numpy import array  # NOQA

    ftype, data = get_file_type(fname, raw_data=True)

    instr = data['instr']
    if ftype == 'lc' and instr.lower() == 'pca':
        logging.warning('RXTE/PCA data; normalizing lc per no. PCUs')
        # If RXTE, plot per PCU count rate
        data['counts'] /= data['nPCUs']
    mjdref = data['mjdref']
    # Map all entries of data to local variables
    locals().update(data)

    good = eval(filter_expr)

    gtis = create_gti_from_condition(locals()['time'], good,
                                     safe_interval=safe_interval)

    gtis = filter_gti_by_length(gtis, minimum_length)

    outfile = _assign_value_if_none(
        outfile, hen_root(fname) + '_gti' + HEN_FILE_EXTENSION)
    save_data({'gti': gtis, 'mjdref': mjdref, '__sr__class__type__': 'gti'},
              outfile)

    return gtis


def apply_gti(fname, gti, outname=None,
              minimum_length=0):
    """Apply a GTI list to the data contained in a file.

    File MUST have a GTI extension already, and an extension called `time`.
    """
    ftype, data = get_file_type(fname, raw_data=True)

    try:
        datagti = data['gti']
        newgtis = cross_gtis([gti, datagti])
    except:  # pragma: no cover
        logging.warning('Data have no GTI extension')
        newgtis = gti

    newgtis = filter_gti_by_length(newgtis, minimum_length)

    data['__sr__class__type__'] = 'gti'
    data['gti'] = newgtis
    good = create_gti_mask(data['time'], newgtis)

    data['time'] = data['time'][good]
    if ftype == 'lc':
        data['counts'] = data['counts'][good]
        data['counts_err'] = data['counts_err'][good]
    elif ftype == 'events':
        data['PI'] = data['PI'][good]
        if data['instr'] == 'PCA':  # pragma: no cover
            data['PCU'] = data['PCU'][good]

    newext = '_gtifilt' + HEN_FILE_EXTENSION
    outname = _assign_value_if_none(
        outname,
        fname.replace(HEN_FILE_EXTENSION, '') + newext)
    save_data(data, outname)

    return newgtis


def main(args=None):
    """Main function called by the `HENcreategti` command line script."""
    import argparse

    description = ('Create GTI files from a filter expression, or applies '
                   'previously created GTIs to a file')
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of files", nargs='+')

    parser.add_argument("-f", "--filter", type=str, default=None,
                        help="Filter expression, that has to be a valid " +
                        "Python boolean operation on a data variable " +
                        "contained in the files")

    parser.add_argument("-c", "--create-only",
                        default=False, action="store_true",
                        help="If specified, creates GTIs withouth applying" +
                        "them to files (Default: False)")

    parser.add_argument("--overwrite",
                        default=False, action="store_true",
                        help="Overwrite original file (Default: False)")

    parser.add_argument("-a", "--apply-gti", type=str, default=None,
                        help="Apply a GTI from this file to input files")

    parser.add_argument("-l", "--minimum-length", type=float, default=0,
                        help=("Minimum length of GTIs (below this length, they"
                              " will be discarded)"))

    parser.add_argument("--safe-interval", nargs=2, type=float,
                        default=[0, 0],
                        help="Interval at start and stop of GTIs used" +
                        " for filtering")

    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING',
                        type=str)

    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')

    args = parser.parse_args(args)
    files = args.files

    if args.debug:
        args.loglevel = 'DEBUG'

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENcreategti.log', level=numeric_level,
                        filemode='w')

    filter_expr = args.filter
    if filter_expr is None and args.apply_gti is None:
        sys.exit("Please specify filter expression (-f option) or input "
                 "GTI file (-a option)")

    for fname in files:
        if args.apply_gti is not None:
            data = load_data(args.apply_gti)
            gtis = data['gti']
        else:
            gtis = create_gti(fname, filter_expr,
                              safe_interval=args.safe_interval,
                              minimum_length=args.minimum_length)
        if args.create_only:
            continue
        if args.overwrite:
            outname = fname
        else:
            # Use default
            outname = None
        apply_gti(fname, gtis, outname=outname,
                  minimum_length=args.minimum_length)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Light curve-related functions."""

from __future__ import (absolute_import, unicode_literals, division,
                        print_function)

import numpy as np
import os
import logging
import warnings
import copy

from stingray.lightcurve import Lightcurve
from stingray.utils import assign_value_if_none
from stingray.gti import create_gti_mask, cross_gtis, contiguous_regions
from .base import _look_for_array_in_array, hen_root, mkdir_p
from .io import load_events, load_data, save_data, save_lcurve, load_lcurve
from .io import HEN_FILE_EXTENSION, high_precision_keyword_read, get_file_type


def join_lightcurves(lcfilelist, outfile='out_lc' + HEN_FILE_EXTENSION):
    """Join light curves from different files.

    Light curves from different instruments are put in different channels.

    Parameters
    ----------
    lcfilelist : list of str
        List of input file names
    outfile :
        Output light curve
    See Also
    --------
        scrunch_lightcurves : Create a single light curve from input light
                                 curves.

    """
    lcdatas = []

    for lfc in lcfilelist:
        logging.info("Loading file %s..." % lfc)
        lcdata = load_lcurve(lfc)
        logging.info("Done.")
        lcdatas.append(lcdata)
        del lcdata

    # --------------- Check consistency of data --------------
    lcdts = [lcdata.dt for lcdata in lcdatas]
    # Find unique elements. If multiple bin times are used, throw an exception
    lcdts = list(set(lcdts))
    assert len(lcdts) == 1, 'Light curves must have same dt for joining'

    instrs = [lcdata.instr for lcdata in lcdatas if hasattr(lcdata, 'instr')]

    # Find unique elements. A lightcurve will be produced for each instrument
    instrs = list(set(instrs))
    if instrs == []:
        instrs = ['unknown']

    outlcs = {}
    for instr in instrs:
        outlcs[instr] = None
    # -------------------------------------------------------

    for lcdata in lcdatas:
        instr = assign_value_if_none(lcdata.instr, 'unknown')
        if outlcs[instr] is None:
            outlcs[instr] = lcdata
        else:
            outlcs[instr] = outlcs[instr].join(lcdata)

    if outfile is not None:
        for instr in instrs:
            if len(instrs) == 1:
                tag = ""
            else:
                tag = instr
            logging.info('Saving joined light curve to %s' % outfile)

            dname, fname = os.path.split(outfile)
            save_lcurve(outlcs[instr], os.path.join(dname, tag + fname))

    return outlcs


def scrunch_lightcurves(lcfilelist, outfile='out_scrlc'+HEN_FILE_EXTENSION,
                        save_joint=False):
    """Create a single light curve from input light curves.

    Light curves are appended when they cover different times, and summed when
    they fall in the same time range. This is done regardless of the channel
    or the instrument.

    Parameters
    ----------
    lcfilelist : list of str
        The list of light curve files to scrunch

    Returns
    -------
    time : array-like
        The time array
    lc :
        The new light curve
    gti : [[gti0_0, gti0_1], [gti1_0, gti1_1], ...]
        Good Time Intervals

    Other Parameters
    ----------------
    outfile : str
        The output file name
    save_joint : bool
        If True, save the per-channel joint light curves

    See Also
    --------
        join_lightcurves : Join light curves from different files
    """
    if save_joint:
        lcdata = join_lightcurves(lcfilelist)
    else:
        lcdata = join_lightcurves(lcfilelist, outfile=None)

    instrs = list(lcdata.keys())
    gti_lists = [lcdata[inst].gti for inst in instrs]
    gti = cross_gtis(gti_lists)
    # Determine limits

    lc0 = lcdata[instrs[0]]

    for inst in instrs[1:]:
        lc0 = lc0 + lcdata[inst]

    lc0.instr = ",".join(instrs)

    logging.info('Saving scrunched light curve to %s' % outfile)
    save_lcurve(lc0, outfile)

    return lc0


def filter_lc_gtis(lc, safe_interval=None, delete=False,
                   min_length=0, return_borders=False):
    """Filter a light curve for GTIs.

    Parameters
    ----------
    lc : :class:`Lightcurve` object
        The input light curve

    Returns
    -------
    newlc : :class:`Lightcurve` object
        The output light curve
    borders : [[i0_0, i0_1], [i1_0, i1_1], ...], optional
        The indexes of the light curve corresponding to the borders of the
        GTIs. Returned if return_borders is set to True

    Other Parameters
    ----------------
    safe_interval : float or [float, float]
        Seconds to filter out at the start and end of each GTI. If single
        float, these safe windows are equal, otherwise the two numbers refer
        to the start and end of the GTI respectively
    delete : bool
        If delete is True, the intervals outside of GTIs are filtered out from
        the light curve. Otherwise, they are set to zero.
    min_length : float
        Minimum length of GTI. GTIs below this length will be removed.
    return_borders : bool
        If True, return also the indexes of the light curve corresponding to
        the borders of the GTIs
    """
    mask, newgtis = create_gti_mask(lc.time, lc.gti,
                                    return_new_gtis=True,
                                    safe_interval=safe_interval,
                                    min_length=min_length)

    nomask = np.logical_not(mask)

    newlc = copy.copy(lc)
    newlc.counts[nomask] = 0
    newlc.gti = newgtis

    if return_borders:
        mask = create_gti_mask(lc.time, newgtis)
        borders = contiguous_regions(mask)
        return newlc, borders
    else:
        return newlc


def lcurve_from_events(f, safe_interval=0,
                       pi_interval=None,
                       e_interval=None,
                       min_length=0,
                       gti_split=False,
                       ignore_gtis=False,
                       bintime=1.,
                       outdir=None,
                       outfile=None,
                       noclobber=False):
    """Bin an event list in a light curve.

    Parameters
    ----------
    f : str
        Input event file name
    bintime : float
        The bin time of the output light curve

    Returns
    -------
    outfiles : list
        List of output light curves

    Other Parameters
    ----------------
    safe_interval : float or [float, float]
        Seconds to filter out at the start and end of each GTI. If single
        float, these safe windows are equal, otherwise the two numbers refer
        to the start and end of the GTI respectively
    pi_interval : [int, int]
        PI channel interval to select. Default None, meaning that all PI
        channels are used
    e_interval : [float, float]
        Energy interval to select (only works if event list is calibrated with
        `calibrate`). Default None
    min_length : float
        GTIs below this length will be filtered out
    gti_split : bool
        If True, create one light curve for each good time interval
    ignore_gtis : bool
        Ignore good time intervals, and get a single light curve that includes
        possible gaps
    outdir : str
        Output directory
    outfile : str
        Output file
    noclobber : bool
        If True, do not overwrite existing files

    """
    logging.info("Loading file %s..." % f)
    evdata = load_events(f)
    logging.info("Done.")

    if bintime < 0:
        bintime = 2 ** (bintime)
    bintime = np.longdouble(bintime)

    tag = ''

    gtis = evdata.gti
    tstart = np.min(gtis)
    tstop = np.max(gtis)
    events = evdata.time
    if hasattr(evdata, 'instr') and evdata.instr is not None:
        instr = evdata.instr
    else:
        instr = "unknown"

    if ignore_gtis:
        gtis = np.array([[tstart, tstop]])
        evdata.gtis = gtis

    total_lc = evdata.to_lc(100)
    total_lc.instr = instr

    # Then, apply filters
    if pi_interval is not None and np.all(np.array(pi_interval) > 0):
        pis = evdata.pi
        good = np.logical_and(pis > pi_interval[0],
                              pis <= pi_interval[1])
        events = events[good]
        tag = '_PI%g-%g' % (pi_interval[0], pi_interval[1])
    elif e_interval is not None and np.all(np.array(e_interval) > 0):
        if not hasattr(evdata, 'energy') or evdata.energy is None:
            raise \
                ValueError("No energy information is present in the file." +
                           " Did you run HENcalibrate?")
        es = evdata.energy
        good = np.logical_and(es > e_interval[0],
                              es <= e_interval[1])
        events = events[good]
        tag = '_E%g-%g' % (e_interval[0], e_interval[1])
    else:
        pass

    if tag != "":
        save_lcurve(total_lc, hen_root(f) + '_std_lc' + HEN_FILE_EXTENSION)

    # Assign default value if None
    outfile = assign_value_if_none(outfile,  hen_root(f) + tag + '_lc')

    # Take out extension from name, if present, then give extension. This
    # avoids multiple extensions
    outfile = outfile.replace(HEN_FILE_EXTENSION, '') + HEN_FILE_EXTENSION
    outdir = assign_value_if_none(
        outdir, os.path.dirname(os.path.abspath(f)))

    _, outfile = os.path.split(outfile)
    mkdir_p(outdir)
    outfile = os.path.join(outdir, outfile)

    if noclobber and os.path.exists(outfile):
        warnings.warn('File exists, and noclobber option used. Skipping')
        return [outfile]

    lc = Lightcurve.make_lightcurve(events, bintime, tstart=tstart,
                                    tseg=tstop-tstart, mjdref=evdata.mjdref,
                                    gti=gtis)

    lc.instr = instr

    lc = filter_lc_gtis(lc, safe_interval=safe_interval,
                        delete=False,
                        min_length=min_length)

    if len(lc.gti) == 0:
        warnings.warn(
            "No GTIs above min_length ({0}s) found.".format(min_length))
        return

    lc.header = None
    if hasattr(evdata, 'header'):
        lc.header = evdata.header

    if gti_split:
        lcs = lc.split_by_gti()
        outfiles = []

        for ib, l0 in enumerate(lcs):
            local_tag = tag + '_gti%d' % ib
            outf = hen_root(outfile) + local_tag + '_lc' + HEN_FILE_EXTENSION
            if noclobber and os.path.exists(outf):
                warnings.warn(
                    'File exists, and noclobber option used. Skipping')
                outfiles.append(outf)
            l0.instr = lc.instr
            l0.header = lc.header

            save_lcurve(l0, outf)
            outfiles.append(outf)
    else:
        logging.info('Saving light curve to %s' % outfile)
        save_lcurve(lc, outfile)
        outfiles = [outfile]

    # For consistency in return value
    return outfiles


def lcurve_from_fits(fits_file, gtistring='GTI',
                     timecolumn='TIME', ratecolumn=None, ratehdu=1,
                     fracexp_limit=0.9, outfile=None,
                     noclobber=False, outdir=None):
    """
    Load a lightcurve from a fits file and save it in HENDRICS format.

    .. note ::
        FITS light curve handling is still under testing.
        Absolute times might be incorrect depending on the light curve format.

    Parameters
    ----------
    fits_file : str
        File name of the input light curve in FITS format

    Returns
    -------
    outfile : [str]
        Returned as a list with a single element for consistency with
        `lcurve_from_events`

    Other Parameters
    ----------------
    gtistring : str
        Name of the GTI extension in the FITS file
    timecolumn : str
        Name of the column containing times in the FITS file
    ratecolumn : str
        Name of the column containing rates in the FITS file
    ratehdu : str or int
        Name or index of the FITS extension containing the light curve
    fracexp_limit : float
        Minimum exposure fraction allowed
    outfile : str
        Output file name
    noclobber : bool
        If True, do not overwrite existing files
    """
    logging.warning(
        """WARNING! FITS light curve handling is still under testing.
        Absolute times might be incorrect.""")
    # TODO:
    # treat consistently TDB, UTC, TAI, etc. This requires some documentation
    # reading. For now, we assume TDB
    from astropy.io import fits as pf
    from astropy.time import Time
    import numpy as np
    from stingray.gti import create_gti_from_condition

    outfile = assign_value_if_none(outfile, hen_root(fits_file) + '_lc')
    outfile = outfile.replace(HEN_FILE_EXTENSION, '') + HEN_FILE_EXTENSION
    outdir = assign_value_if_none(
        outdir, os.path.dirname(os.path.abspath(fits_file)))

    _, outfile = os.path.split(outfile)
    mkdir_p(outdir)
    outfile = os.path.join(outdir, outfile)

    if noclobber and os.path.exists(outfile):
        warnings.warn('File exists, and noclobber option used. Skipping')
        return [outfile]

    lchdulist = pf.open(fits_file)
    lctable = lchdulist[ratehdu].data

    # Units of header keywords
    tunit = lchdulist[ratehdu].header['TIMEUNIT']

    try:
        mjdref = high_precision_keyword_read(lchdulist[ratehdu].header,
                                             'MJDREF')
        mjdref = Time(mjdref, scale='tdb', format='mjd')
    except:
        mjdref = None

    try:
        instr = lchdulist[ratehdu].header['INSTRUME']
    except:
        instr = 'EXTERN'

    # ----------------------------------------------------------------
    # Trying to comply with all different formats of fits light curves.
    # It's a madness...
    try:
        tstart = high_precision_keyword_read(lchdulist[ratehdu].header,
                                             'TSTART')
        tstop = high_precision_keyword_read(lchdulist[ratehdu].header,
                                            'TSTOP')
    except:
        raise(Exception('TSTART and TSTOP need to be specified'))

    # For nulccorr lcs this whould work

    timezero = high_precision_keyword_read(lchdulist[ratehdu].header,
                                           'TIMEZERO')
    # Sometimes timezero is "from tstart", sometimes it's an absolute time.
    # This tries to detect which case is this, and always consider it
    # referred to tstart
    timezero = assign_value_if_none(timezero, 0)

    # for lcurve light curves this should instead work
    if tunit == 'd':
        # TODO:
        # Check this. For now, I assume TD (JD - 2440000.5).
        # This is likely wrong
        timezero = Time(2440000.5 + timezero, scale='tdb', format='jd')
        tstart = Time(2440000.5 + tstart, scale='tdb', format='jd')
        tstop = Time(2440000.5 + tstop, scale='tdb', format='jd')
        # if None, use NuSTAR defaulf MJDREF
        mjdref = assign_value_if_none(
            mjdref, Time(np.longdouble('55197.00076601852'), scale='tdb',
                         format='mjd'))

        timezero = (timezero - mjdref).to('s').value
        tstart = (tstart - mjdref).to('s').value
        tstop = (tstop - mjdref).to('s').value

    if timezero > tstart:
        timezero -= tstart

    time = np.array(lctable.field(timecolumn), dtype=np.longdouble)
    if time[-1] < tstart:
        time += timezero + tstart
    else:
        time += timezero

    try:
        dt = high_precision_keyword_read(lchdulist[ratehdu].header,
                                         'TIMEDEL')
        if tunit == 'd':
            dt *= 86400
    except:
        warnings.warn('Assuming that TIMEDEL is the difference between the'
                      ' first two times of the light curve')
        dt = time[1] - time[0]

    # ----------------------------------------------------------------
    ratecolumn = assign_value_if_none(
        ratecolumn,
        _look_for_array_in_array(['RATE', 'RATE1', 'COUNTS'], lctable.names))

    rate = np.array(lctable.field(ratecolumn), dtype=np.float)

    try:
        rate_e = np.array(lctable.field('ERROR'), dtype=np.longdouble)
    except:
        rate_e = np.zeros_like(rate)

    if 'RATE' in ratecolumn:
        rate *= dt
        rate_e *= dt

    try:
        fracexp = np.array(lctable.field('FRACEXP'), dtype=np.longdouble)
    except:
        fracexp = np.ones_like(rate)

    good_intervals = (rate == rate) * (fracexp >= fracexp_limit) * \
        (fracexp <= 1)

    rate[good_intervals] /= fracexp[good_intervals]
    rate_e[good_intervals] /= fracexp[good_intervals]

    rate[np.logical_not(good_intervals)] = 0

    try:
        gtitable = lchdulist[gtistring].data
        gti_list = np.array([[a, b]
                             for a, b in zip(gtitable.field('START'),
                                             gtitable.field('STOP'))],
                            dtype=np.longdouble)
    except:
        gti_list = create_gti_from_condition(time, good_intervals)

    lchdulist.close()

    lc = Lightcurve(time=time, counts=rate, err=rate_e, gti=gti_list,
                    mjdref=mjdref.mjd)

    lc.instr = instr
    lc.header = lchdulist[ratehdu].header.tostring()

    logging.info('Saving light curve to %s' % outfile)
    save_lcurve(lc, outfile)
    return [outfile]


def lcurve_from_txt(txt_file, outfile=None,
                    noclobber=False, outdir=None,
                    mjdref=None, gti=None):
    """
    Load a lightcurve from a text file.

    Parameters
    ----------
    txt_file : str
        File name of the input light curve in text format. Assumes two columns:
        time, counts. Times are seconds from MJDREF 55197.00076601852 (NuSTAR)
        if not otherwise specified.

    Returns
    -------
    outfile : [str]
        Returned as a list with a single element for consistency with
        `lcurve_from_events`

    Other Parameters
    ----------------
    outfile : str
        Output file name
    noclobber : bool
        If True, do not overwrite existing files
    mjdref : float, default 55197.00076601852
        the MJD time reference
    gti : [[gti0_0, gti0_1], [gti1_0, gti1_1], ...]
        Good Time Intervals
    """
    import numpy as np
    if mjdref is None:
        mjdref = np.longdouble('55197.00076601852')

    outfile = assign_value_if_none(outfile, hen_root(txt_file) + '_lc')
    outfile = outfile.replace(HEN_FILE_EXTENSION, '') + HEN_FILE_EXTENSION

    outdir = assign_value_if_none(
        outdir, os.path.dirname(os.path.abspath(txt_file)))

    _, outfile = os.path.split(outfile)
    mkdir_p(outdir)
    outfile = os.path.join(outdir, outfile)

    if noclobber and os.path.exists(outfile):
        warnings.warn('File exists, and noclobber option used. Skipping')
        return [outfile]

    time, counts = np.genfromtxt(txt_file, delimiter=' ', unpack=True)
    time = np.array(time, dtype=np.longdouble)
    counts = np.array(counts, dtype=np.float)

    lc = Lightcurve(time=time, counts=counts, gti=gti,
                    mjdref=mjdref)

    lc.instr = 'EXTERN'

    logging.info('Saving light curve to %s' % outfile)
    save_lcurve(lc, outfile)
    return [outfile]


def _baseline_lightcurves(lcurves, outroot, p, lam):
    outroot_save = outroot
    for i, f in enumerate(lcurves):
        if outroot is None:
            outroot = hen_root(f) + '_lc_baseline'
        else:
            outroot = outroot_save + '_{}'.format(i)
        ftype, lc = get_file_type(f)
        baseline = lc.baseline(p, lam)
        lc.base = baseline
        save_lcurve(lc, outroot + HEN_FILE_EXTENSION)


def _wrap_lc(args):
    f, kwargs = args
    try:
        return lcurve_from_events(f, **kwargs)
    except Exception as e:
        warnings.warn("HENlcurve exception: {0}".format(str(e)))
        raise
        return []


def _wrap_txt(args):
    f, kwargs = args
    try:
        return lcurve_from_txt(f, **kwargs)
    except Exception as e:
        warnings.warn("HENlcurve exception: {0}".format(str(e)))
        return []


def _wrap_fits(args):
    f, kwargs = args
    try:
        return lcurve_from_fits(f, **kwargs)
    except Exception as e:
        warnings.warn("HENlcurve exception: {0}".format(str(e)))
        return []


def main(args=None):
    """Main function called by the `HENlcurve` command line script."""
    import argparse
    from multiprocessing import Pool

    description = ('Create lightcurves starting from event files. It is '
                   'possible to specify energy or channel filtering options')
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of files", nargs='+')

    parser.add_argument("-b", "--bintime", type=float, default=1,
                        help="Bin time; if negative, negative power of 2")
    parser.add_argument("--safe-interval", nargs=2, type=float,
                        default=[0, 0],
                        help="Interval at start and stop of GTIs used" +
                        " for filtering")
    parser.add_argument("--pi-interval", type=int, default=[-1, -1],
                        nargs=2,
                        help="PI interval used for filtering")
    parser.add_argument('-e', "--e-interval", type=float, default=[-1, -1],
                        nargs=2,
                        help="Energy interval used for filtering")
    parser.add_argument("-s", "--scrunch",
                        help="Create scrunched light curve (single channel)",
                        default=False,
                        action="store_true")
    parser.add_argument("-j", "--join",
                        help="Create joint light curve (multiple channels)",
                        default=False,
                        action="store_true")
    parser.add_argument("-g", "--gti-split",
                        help="Split light curve by GTI",
                        default=False,
                        action="store_true")
    parser.add_argument("--minlen",
                        help="Minimum length of acceptable GTIs (default:4)",
                        default=4, type=float)
    parser.add_argument("--ignore-gtis",
                        help="Ignore GTIs",
                        default=False,
                        action="store_true")
    parser.add_argument("-d", "--outdir", type=str, default=None,
                        help='Output directory')
    parser.add_argument("-o", "--outfile", type=str, default=None,
                        help='Output file name')
    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING',
                        type=str)
    parser.add_argument("--nproc",
                        help=("Number of processors to use"),
                        default=1,
                        type=int)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')
    parser.add_argument("--noclobber", help="Do not overwrite existing files",
                        default=False, action='store_true')
    parser.add_argument("--fits-input",
                        help="Input files are light curves in FITS format",
                        default=False, action='store_true')
    parser.add_argument("--txt-input",
                        help="Input files are light curves in txt format",
                        default=False, action='store_true')

    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = 'DEBUG'
    bintime = args.bintime

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENlcurve.log', level=numeric_level,
                        filemode='w')

    safe_interval = args.safe_interval
    pi_interval = np.array(args.pi_interval)
    e_interval = np.array(args.e_interval)

    # ------ Use functools.partial to wrap lcurve* with relevant keywords---
    if args.fits_input:
        wrap_fun = _wrap_fits
        argdict = {"noclobber": args.noclobber}
    elif args.txt_input:
        wrap_fun = _wrap_txt
        argdict = {"noclobber": args.noclobber}
    else:
        wrap_fun = _wrap_lc
        argdict = {"noclobber": args.noclobber, "safe_interval": safe_interval,
                   "pi_interval": pi_interval,
                   "e_interval": e_interval,
                   "min_length": args.minlen,
                   "gti_split": args.gti_split,
                   "ignore_gtis": args.ignore_gtis,
                   "bintime": bintime, "outdir": args.outdir}

    arglist = [[f, argdict.copy()] for f in args.files]
    na = len(arglist)
    outfile = args.outfile
    if outfile is not None:
        outname = os.path.splitext(outfile)[0]
        for i in range(na):
            if na > 1:
                outname = outfile + "_{0}".format(i)
            arglist[i][1]['outfile'] = outname

    # -------------------------------------------------------------------------
    outfiles = []

    if os.name == 'nt' or args.nproc == 1:
        for a in arglist:
            outfiles.append(wrap_fun(a))
    else:
        pool = Pool(processes=args.nproc)
        for i in pool.imap_unordered(wrap_fun, arglist):
            outfiles.append(i)
        pool.close()

    logging.debug(outfiles)

    if args.scrunch:
        scrunch_lightcurves(outfiles)

    if args.join:
        join_lightcurves(outfiles)


def scrunch_main(args=None):
    """Main function called by the `HENscrunchlc` command line script."""
    import argparse
    description = \
        'Sum lightcurves from different instruments or energy ranges'
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="List of files", nargs='+')
    parser.add_argument("-o", "--out", type=str,
                        default="out_scrlc" + HEN_FILE_EXTENSION,
                        help='Output file')
    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING',
                        type=str)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')

    args = parser.parse_args(args)
    files = args.files

    if args.debug:
        args.loglevel = 'DEBUG'

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENscrunchlc.log', level=numeric_level,
                        filemode='w')

    scrunch_lightcurves(files, args.out)


def baseline_main(args=None):
    """Main function called by the `HENbaselinesub` command line script."""
    import argparse
    description = \
        ('Subtract a baseline from the lightcurve using the Asymmetric Least '
         'Squares algorithm. The two parameters p and lambda control the '
         'asymmetry and smoothness of the baseline. See below for details.')
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="List of files", nargs='+')
    parser.add_argument("-o", "--out", type=str, default=None,
                        help='Output file')
    parser.add_argument("--loglevel",
                        help=("use given logging level (one between INFO, "
                              "WARNING, ERROR, CRITICAL, DEBUG; "
                              "default:WARNING)"),
                        default='WARNING',
                        type=str)
    parser.add_argument("--debug", help="use DEBUG logging level",
                        default=False, action='store_true')
    parser.add_argument("-p", '--asymmetry', type=float,
                        help='"asymmetry" parameter. Smaller values make the '
                             'baseline more "horizontal". Typically '
                             '0.001 < p < 0.1, but not necessarily.',
                        default=0.01)
    parser.add_argument("-l", "--lam", type=float,
                        help='lambda, or "smoothness", parameter. Larger'
                        ' values make the baseline stiffer. Typically '
                        '1e2 < lam < 1e9',
                        default=1e5)

    args = parser.parse_args(args)
    files = args.files

    if args.debug:
        args.loglevel = 'DEBUG'

    numeric_level = getattr(logging, args.loglevel.upper(), None)
    logging.basicConfig(filename='HENscrunchlc.log', level=numeric_level,
                        filemode='w')

    _baseline_lightcurves(files, args.out, args.asymmetry, args.lam)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test a full run of the codes from the command line."""

from __future__ import (absolute_import, unicode_literals, division,
                        print_function)

import hendrics as hen
import logging
import os
import glob
import subprocess as sp
import numpy as np
from astropy.tests.helper import catch_warnings
from astropy.io import fits
import pytest
from stingray.lightcurve import Lightcurve
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

HEN_FILE_EXTENSION = hen.io.HEN_FILE_EXTENSION

logging.basicConfig(filename='HEN.log', level=logging.DEBUG, filemode='w')


class TestFullRun(object):
    """Test how command lines work.

    Usually considered bad practice, but in this
    case I need to test the full run of the codes, and files depend on each
    other.
    Inspired by http://stackoverflow.com/questions/5387299/python-unittest-testcase-execution-order

    When command line is missing, uses some function calls
    """  # NOQA
    @classmethod
    def setup_class(cls):
        curdir = os.path.abspath(os.path.dirname(__file__))
        cls.datadir = os.path.join(curdir, 'data')
        cls.first_event_file = os.path.join(cls.datadir,
                                            'monol_testA_nustar_fpma_ev' +
                                            HEN_FILE_EXTENSION)

    def test_scripts_are_installed(self):
        """Test only once that command line scripts are installed correctly."""
        fits_file = os.path.join(self.datadir, 'monol_testA.evt')
        command = 'HENreadfile {0}'.format(fits_file)
        sp.check_call(command.split())

    def test_fake_file(self):
        """Test produce a fake event file."""
        fits_file = os.path.join(self.datadir, 'monol_test_fake.evt')
        hen.fake.main(['-o', fits_file, '--instrument', 'FPMB'])
        info = hen.io.print_fits_info(fits_file, hdu=1)
        assert info['Instrument'] == 'FPMB'

    def test_fake_file_from_input_lc(self):
        """Test produce a fake event file from input light curve."""
        lcurve_in = os.path.join(self.datadir, 'lcurveA.fits')
        fits_file = os.path.join(self.datadir, 'monol_test_fake_lc.evt')
        hen.fake.main(['--lc', lcurve_in, '-o', fits_file])

    def test_fake_file_with_deadtime(self):
        """Test produce a fake event file and apply deadtime."""
        fits_file = os.path.join(self.datadir, 'monol_test_fake_lc.evt')
        hen.fake.main(['--deadtime', '2.5e-3',
                       '--ctrate', '2000',
                       '-o', fits_file])

    def test_fake_file_xmm(self):
        """Test produce a fake event file and apply deadtime."""
        fits_file = os.path.join(self.datadir, 'monol_test_fake_lc_xmm.evt')
        hen.fake.main(['--deadtime', '1e-4', '-m', 'XMM', '-i', 'epn',
                       '--ctrate', '2000',
                       '-o', fits_file])
        hdu_list = fits.open(fits_file)
        hdunames = [hdu.name for hdu in hdu_list]
        assert 'STDGTI01' in hdunames
        assert 'STDGTI02' in hdunames
        assert 'STDGTI07' in hdunames

    def test_load_events(self):
        """Test event file reading."""
        command = '{0}'.format(
            os.path.join(self.datadir, 'monol_testA.evt'))
        hen.read_events.main(command.split())
        new_filename = self.first_event_file
        ev = hen.io.load_events(new_filename)
        assert hasattr(ev, 'header')
        assert hasattr(ev, 'gti')

    def test_load_events_with_2_cpus(self):
        """Test event file reading."""
        command = '{0} {1} --nproc 2'.format(
            os.path.join(self.datadir, 'monol_testB.evt'),
            os.path.join(self.datadir, 'monol_testA_timezero.evt'),
            os.path.join(self.datadir, 'monol_test_fake.evt'))
        hen.read_events.main(command.split())

    def test_load_events_split(self):
        """Test event file splitting."""
        command = \
            '{0} -g --min-length 0'.format(
                os.path.join(self.datadir, 'monol_testB.evt'))
        hen.read_events.main(command.split())
        new_filename = os.path.join(self.datadir,
                                    'monol_testB_nustar_fpmb_gti0_ev' +
                                    HEN_FILE_EXTENSION)
        assert os.path.exists(new_filename)

    def test_save_binary_events(self):
        f = self.first_event_file
        with pytest.raises(ValueError) as excinfo:
            hen.binary.main_presto("{} -b 0.1 -e 3 59".format(f).split())

        assert 'Energy filtering requested' in str(excinfo)

    def test_load_gtis(self):
        """Test loading of GTIs from FITS files."""
        fits_file = os.path.join(self.datadir, 'monol_testA.evt')
        hen.io.load_gtis(fits_file)

    def test_load_events_noclobber(self):
        """Test event file reading w. noclobber option."""
        with catch_warnings() as w:
            command = \
                '{0} --noclobber'.format(
                    os.path.join(self.datadir, 'monol_testB.evt'))
            hen.read_events.main(command.split())
        assert str(w[0].message).strip().endswith(
            "exists and using noclobber. Skipping"), \
            "Unexpected warning output"

    def test_load_events_xmm(self):
        """Test event file reading."""
        command = '{0}'.format(
            os.path.join(self.datadir, 'monol_test_fake_lc_xmm.evt'))
        hen.read_events.main(command.split())

    def test_calibrate(self):
        """Test event file calibration."""
        from astropy.io.fits import Header
        command = '{0} -r {1}'.format(
            os.path.join(self.datadir,
                         'monol_testA_nustar_fpma_ev' + HEN_FILE_EXTENSION),
            os.path.join(self.datadir, 'test.rmf'))
        hen.calibrate.main(command.split())
        new_filename = os.path.join(self.datadir,
                                    'monol_testA_nustar_fpma_ev_calib' +
                                    HEN_FILE_EXTENSION)
        assert os.path.exists(new_filename)
        ev = hen.io.load_events(new_filename)
        assert hasattr(ev, 'header')

        Header.fromstring(ev.header)
        assert hasattr(ev, 'gti')
        gti_to_test = hen.io.load_events(self.first_event_file).gti
        assert np.allclose(gti_to_test, ev.gti)

    def test_save_binary_calibrated_events(self):
        f = os.path.join(self.datadir,
                         'monol_testA_nustar_fpma_ev_calib' +
                         HEN_FILE_EXTENSION)
        hen.binary.main_presto("{} -b 0.1 -e 3 59".format(f).split())
        assert os.path.exists(f.replace(HEN_FILE_EXTENSION, '.dat'))
        assert os.path.exists(f.replace(HEN_FILE_EXTENSION, '.inf'))

    def test_calibrate_2_cpus(self):
        """Test event file calibration."""
        command = '{0} -r {1} --nproc 2'.format(
            os.path.join(self.datadir,
                         'monol_testB_nustar_fpmb_ev' + HEN_FILE_EXTENSION),
            os.path.join(self.datadir, 'test.rmf'))
        hen.calibrate.main(command.split())
        assert os.path.exists(os.path.join(self.datadir,
                                           'monol_testB_nustar_fpmb_ev_calib' +
                                           HEN_FILE_EXTENSION))

    def test_save_varen_rms(self):
        fname = os.path.join(self.datadir,
                             'monol_testA_nustar_fpma_ev_calib' +
                             HEN_FILE_EXTENSION)
        hen.varenergy.main([fname, "-f", "0", "100", "--energy-values",
                            "0.3", "12", "5", "lin", "--rms", "-b", "0.5",
                            "--segment-size", "128"])
        out = hen.base.hen_root(fname) + "_rms" + '.qdp'
        os.path.exists(out)

    def test_save_varen_lag(self):
        fname = os.path.join(self.datadir,
                             'monol_testA_nustar_fpma_ev_calib' +
                             HEN_FILE_EXTENSION)
        hen.varenergy.main([fname, "-f", "0", "100", "--energy-values",
                            "0.3", "12", "5", "lin", "--lag", "-b", "0.5",
                            "--segment-size", "128"])
        out = hen.base.hen_root(fname) + "_lag" + '.qdp'
        os.path.exists(out)

    def test_lcurve(self):
        """Test light curve production."""
        from astropy.io.fits import Header
        command = ('{0} -e {1} {2} --safe-interval '
                   '{3} {4}  --nproc 2 -b 0.5 -o {5}').format(
            os.path.join(self.datadir, 'monol_testA_nustar_fpma_ev_calib' +
                         HEN_FILE_EXTENSION),
            3, 50, 100, 300,
            os.path.join(self.datadir, 'monol_testA_E3-50_lc' +
                         HEN_FILE_EXTENSION)
        )
        hen.lcurve.main(command.split())

        new_filename = \
            os.path.join(os.path.join(self.datadir,
                                      'monol_testA_E3-50_lc' +
                                      HEN_FILE_EXTENSION))
        assert os.path.exists(new_filename)
        lc = hen.io.load_lcurve(new_filename)
        assert hasattr(lc, 'header')
        # Test that the header is correctly conserved
        Header.fromstring(lc.header)
        assert hasattr(lc, 'gti')
        gti_to_test = hen.io.load_events(self.first_event_file).gti
        assert np.allclose(gti_to_test, lc.gti)

    def test_save_binary_lc(self):
        f = \
            os.path.join(os.path.join(self.datadir,
                                      'monol_testA_E3-50_lc' +
                                      HEN_FILE_EXTENSION))
        hen.binary.main_presto("{}".format(f).split())
        assert os.path.exists(f.replace(HEN_FILE_EXTENSION, '.dat'))
        assert os.path.exists(f.replace(HEN_FILE_EXTENSION, '.inf'))

    def test_lcurve_B(self):
        command = ('{0} -e {1} {2} --safe-interval '
                   '{3} {4} -b 0.5 -o {5}').format(
            os.path.join(self.datadir, 'monol_testB_nustar_fpmb_ev_calib' +
                         HEN_FILE_EXTENSION),
            3, 50, 100, 300,
            os.path.join(self.datadir, 'monol_testB_E3-50_lc' +
                         HEN_FILE_EXTENSION))
        hen.lcurve.main(command.split())
        assert os.path.exists(os.path.join(self.datadir,
                                           'monol_testB_E3-50_lc' +
                                           HEN_FILE_EXTENSION))

    def test_lcurve_from_split_event(self):
        """Test lc reading of split event file."""
        command = '{0}'.format(
            os.path.join(self.datadir, 'monol_testB_nustar_fpmb_gti0_ev' +
                         HEN_FILE_EXTENSION))
        hen.lcurve.main(command.split())
        new_filename = os.path.join(self.datadir,
                                    'monol_testB_nustar_fpmb_gti0_lc' +
                                    HEN_FILE_EXTENSION)
        assert os.path.exists(new_filename)
        lc = hen.io.load_lcurve(new_filename)
        gti_to_test = hen.io.load_events(self.first_event_file).gti[0]
        assert np.allclose(gti_to_test, lc.gti)

    def test_lcurve_split(self):
        """Test lc with gti-split option."""
        command = '{0} {1} -g'.format(
            os.path.join(self.datadir, 'monol_testA_nustar_fpma_ev' +
                         HEN_FILE_EXTENSION),
            os.path.join(self.datadir, 'monol_testB_nustar_fpmb_ev' +
                         HEN_FILE_EXTENSION))
        hen.lcurve.main(command.split())
        new_filename = os.path.join(self.datadir,
                                    'monol_testA_nustar_fpma_gti0_lc' +
                                    HEN_FILE_EXTENSION)
        assert os.path.exists(new_filename)
        lc = hen.io.load_lcurve(new_filename)
        gti_to_test = hen.io.load_events(self.first_event_file).gti[0]
        assert np.allclose(gti_to_test, lc.gti)

    def test_fits_lcurve0(self):
        """Test light curves from FITS."""
        lcurve_ftools_orig = os.path.join(self.datadir, 'lcurveA.fits')

        lcurve_ftools = os.path.join(self.datadir,
                                     'lcurve_ftools_lc' +
                                     HEN_FILE_EXTENSION)

        command = "{0} --outfile {1}".format(
            os.path.join(self.datadir,
                         'monol_testA_nustar_fpma_ev') + HEN_FILE_EXTENSION,
            os.path.join(self.datadir,
                         'lcurve_lc'))
        hen.lcurve.main(command.split())
        assert os.path.exists(os.path.join(self.datadir,
                              'lcurve_lc') + HEN_FILE_EXTENSION)

        command = "--fits-input {0} --outfile {1}".format(
            lcurve_ftools_orig,
            lcurve_ftools)
        hen.lcurve.main(command.split())

    def test_fits_lcurve1(self):
        """Test light curves from FITS."""
        lcurve_ftools = os.path.join(self.datadir,
                                     'lcurve_ftools_lc' +
                                     HEN_FILE_EXTENSION)

        lcurve_mp = os.path.join(self.datadir,
                                 'lcurve_lc' +
                                 HEN_FILE_EXTENSION)

        lcdata_mp = hen.io.load_data(lcurve_mp)
        lcdata_ftools = hen.io.load_data(lcurve_ftools)

        lc_mp = lcdata_mp['counts']

        lenmp = len(lc_mp)
        lc_ftools = lcdata_ftools['counts']
        lenftools = len(lc_ftools)
        goodlen = min([lenftools, lenmp])

        diff = lc_mp[:goodlen] - lc_ftools[:goodlen]

        assert np.all(np.abs(diff) <= 1e-3), \
            'Light curve data do not coincide between FITS and HEN'

    def test_txt_lcurve(self):
        """Test light curves from txt."""
        lcurve_mp = os.path.join(self.datadir,
                                 'lcurve_lc' +
                                 HEN_FILE_EXTENSION)
        lcdata_mp = hen.io.load_data(lcurve_mp)
        lc_mp = lcdata_mp['counts']
        time_mp = lcdata_mp['time']

        lcurve_txt_orig = os.path.join(self.datadir,
                                       'lcurve_txt_lc.txt')

        hen.io.save_as_ascii([time_mp, lc_mp], lcurve_txt_orig)

        lcurve_txt = os.path.join(self.datadir,
                                  'lcurve_txt_lc' +
                                  HEN_FILE_EXTENSION)
        hen.lcurve.main(['--txt-input', lcurve_txt_orig,
                        '--outfile', lcurve_txt])
        lcdata_txt = hen.io.load_data(lcurve_txt)

        lc_txt = lcdata_txt['counts']

        assert np.all(np.abs(lc_mp - lc_txt) <= 1e-3), \
            'Light curve data do not coincide between txt and HEN'

    def test_joinlcs(self):
        """Test produce joined light curves."""
        new_filename = os.path.join(
            self.datadir, 'monol_test_joinlc' + HEN_FILE_EXTENSION)
        # because join_lightcurves separates by instrument
        new_actual_filename = os.path.join(
            self.datadir, 'FPMAmonol_test_joinlc' + HEN_FILE_EXTENSION)
        hen.lcurve.join_lightcurves(
            glob.glob(os.path.join(self.datadir,
                                   'monol_testA_nustar_fpma_gti[0-9]_lc*')) +
            glob.glob(os.path.join(self.datadir,
                                   'monol_testB_nustar_fpmb_gti[0-9]_lc*')),
            new_filename)

        lc = hen.io.load_lcurve(new_actual_filename)
        assert hasattr(lc, 'gti')
        gti_to_test = hen.io.load_events(self.first_event_file).gti
        assert np.allclose(gti_to_test, lc.gti)

    def test_scrunchlcs(self):
        """Test produce scrunched light curves."""
        a_in = os.path.join(self.datadir,
                            'monol_testA_E3-50_lc' + HEN_FILE_EXTENSION)
        b_in = os.path.join(self.datadir,
                            'monol_testB_E3-50_lc' + HEN_FILE_EXTENSION)
        out = os.path.join(self.datadir,
                           'monol_test_scrunchlc' + HEN_FILE_EXTENSION)
        command = '{0} {1} -o {2}'.format(a_in, b_in, out)

        hen.lcurve.scrunch_main(command.split())
        a_lc = hen.io.load_lcurve(a_in)
        b_lc = hen.io.load_lcurve(b_in)
        out_lc = hen.io.load_lcurve(out)
        assert np.all(out_lc.counts == a_lc.counts + b_lc.counts)
        gti_to_test = hen.io.load_events(self.first_event_file).gti
        assert np.allclose(gti_to_test, out_lc.gti)

    def testbaselinelc(self):
        """Test produce scrunched light curves."""
        a_in = os.path.join(self.datadir,
                            'monol_testA_E3-50_lc' + HEN_FILE_EXTENSION)
        out = os.path.join(self.datadir, 'monol_test_baselc')
        command = '{0} -o {1} -p 0.001 --lam 1e5'.format(a_in, out)

        hen.lcurve.baseline_main(command.split())
        out_lc = hen.io.load_lcurve(out + '_0' + HEN_FILE_EXTENSION)
        assert hasattr(out_lc, 'base')
        gti_to_test = hen.io.load_events(self.first_event_file).gti
        assert np.allclose(gti_to_test, out_lc.gti)

    def testbaselinelc_nooutroot(self):
        """Test produce scrunched light curves."""
        a_in = os.path.join(self.datadir,
                            'monol_testA_E3-50_lc' + HEN_FILE_EXTENSION)
        command = '{0} -p 0.001 --lam 1e5'.format(a_in)

        hen.lcurve.baseline_main(command.split())
        out_lc = hen.io.load_lcurve(
            hen.base.hen_root(a_in) + '_lc_baseline' + HEN_FILE_EXTENSION)
        assert hasattr(out_lc, 'base')
        gti_to_test = hen.io.load_events(self.first_event_file).gti
        assert np.allclose(gti_to_test, out_lc.gti)

    def test_lcurve_error_uncalibrated(self):
        """Test light curve error from uncalibrated file."""
        command = ('{0} -e {1} {2}').format(
            os.path.join(self.datadir,
                         'monol_testA_nustar_fpma_ev' + HEN_FILE_EXTENSION),
            3, 50)

        with pytest.raises(ValueError) as excinfo:
            hen.lcurve.main(command.split())
        message = str(excinfo.value)
        assert str(message).strip().endswith("Did you run HENcalibrate?")

    def test_lcurve_pi_filtering(self):
        """Test light curve using PI filtering."""
        command = ('{0} --pi-interval {1} {2}').format(
            os.path.join(self.datadir,
                         'monol_testA_nustar_fpma_ev' + HEN_FILE_EXTENSION),
            10, 300)

        hen.lcurve.main(command.split())

    def test_colors_fail_uncalibrated(self):
        """Test light curve using PI filtering."""
        command = ('{0} -b 100 -e {1} {2} {2} {3}').format(
            os.path.join(self.datadir,
                         'monol_testA_nustar_fpma_ev' + HEN_FILE_EXTENSION),
            3, 5, 10)
        with pytest.raises(ValueError) as excinfo:
            hen.colors.main(command.split())

        assert "No energy information is present " in str(excinfo.value)

    def test_colors(self):
        """Test light curve using PI filtering."""
        # calculate colors
        command = ('{0} -b 100 -e {1} {2} {2} {3}').format(
            os.path.join(self.datadir,
                         'monol_testA_nustar_fpma_ev_calib' +
                         HEN_FILE_EXTENSION),
            3, 5, 10)
        hen.colors.main(command.split())

        new_filename = \
            os.path.join(self.datadir,
                         'monol_testA_nustar_fpma_E_10-5_over_5-3' +
                         HEN_FILE_EXTENSION)
        assert os.path.exists(new_filename)
        out_lc = hen.io.load_lcurve(new_filename)
        gti_to_test = hen.io.load_events(self.first_event_file).gti
        assert np.allclose(gti_to_test, out_lc.gti)

    def test_pds_leahy(self):
        """Test PDS production."""
        lc = os.path.join(self.datadir,
                          'monol_testA_E3-50_lc') + HEN_FILE_EXTENSION
        hen.io.main([lc])
        command = \
            '{0} -f 128 -k PDS --norm leahy'.format(lc)
        hen.fspec.main(command.split())

        assert os.path.exists(
            os.path.join(self.datadir,
                         'monol_testA_E3-50_pds' + HEN_FILE_EXTENSION))

    def test_pds(self):
        """Test PDS production."""
        command = \
            '{0} {1} -f 128 --save-dyn -k PDS --norm frac --nproc 2 '.format(
                os.path.join(self.datadir,
                             'monol_testA_E3-50_lc') + HEN_FILE_EXTENSION,
                os.path.join(self.datadir,
                             'monol_testB_E3-50_lc') + HEN_FILE_EXTENSION)
        hen.fspec.main(command.split())

        assert os.path.exists(
            os.path.join(self.datadir,
                         'monol_testB_E3-50_pds' + HEN_FILE_EXTENSION))
        assert os.path.exists(
            os.path.join(self.datadir,
                         'monol_testA_E3-50_pds') + HEN_FILE_EXTENSION)

    def test_pds_fits(self):
        """Test PDS production with light curves obtained from FITS files."""
        lcurve_ftools = os.path.join(self.datadir,
                                     'lcurve_ftools_lc' +
                                     HEN_FILE_EXTENSION)
        command = '{0} -f 128'.format(lcurve_ftools)
        hen.fspec.main(command.split())

    def test_pds_txt(self):
        """Test PDS production with light curves obtained from txt files."""
        lcurve_txt = os.path.join(self.datadir,
                                  'lcurve_txt_lc' +
                                  HEN_FILE_EXTENSION)
        command = '{0} -f 128'.format(lcurve_txt)
        hen.fspec.main(command.split())

    def test_cpds_rms_norm(self):
        """Test CPDS production."""
        command = \
            '{0} {1} -f 128 --save-dyn -k CPDS --norm rms -o {2}'.format(
                os.path.join(self.datadir, 'monol_testA_E3-50_lc') +
                HEN_FILE_EXTENSION,
                os.path.join(self.datadir, 'monol_testB_E3-50_lc') +
                HEN_FILE_EXTENSION,
                os.path.join(self.datadir, 'monol_test_E3-50'))

        hen.fspec.main(command.split())

    def test_cpds_wrong_norm(self):
        """Test CPDS production."""
        command = \
            '{0} {1} -f 128 --save-dyn -k CPDS --norm blablabla -o {2}'.format(
                os.path.join(self.datadir, 'monol_testA_E3-50_lc') +
                HEN_FILE_EXTENSION,
                os.path.join(self.datadir, 'monol_testB_E3-50_lc') +
                HEN_FILE_EXTENSION,
                os.path.join(self.datadir, 'monol_test_E3-50'))
        with pytest.warns(UserWarning) as record:
            hen.fspec.main(command.split())

        assert np.any(["Beware! Unknown normalization" in r.message.args[0]
                       for r in record])

    def test_cpds(self):
        """Test CPDS production."""
        command = \
            '{0} {1} -f 128 --save-dyn -k CPDS --norm frac -o {2}'.format(
                os.path.join(self.datadir, 'monol_testA_E3-50_lc') +
                HEN_FILE_EXTENSION,
                os.path.join(self.datadir, 'monol_testB_E3-50_lc') +
                HEN_FILE_EXTENSION,
                os.path.join(self.datadir, 'monol_test_E3-50'))
        hen.fspec.main(command.split())

    def test_cpds_2cpus(self):
        """Test CPDS production."""
        command = \
            ('{0} {1} -f 128 --save-dyn -k '
             'CPDS --norm frac -o {2} --nproc 2').format(
                os.path.join(self.datadir, 'monol_testA_E3-50_lc') +
                HEN_FILE_EXTENSION,
                os.path.join(self.datadir, 'monol_testB_E3-50_lc') +
                HEN_FILE_EXTENSION,
                os.path.join(self.datadir, 'monol_test_E3-50'))
        hen.fspec.main(command.split())

    # def test_dumpdynpds(self):
    #     """Test dump dynamical PDSs."""
    #     command = '--noplot ' + \
    #         os.path.join(self.datadir,
    #                      'monol_testA_E3-50_pds') + \
    #         HEN_FILE_EXTENSION
    #     hen.fspec.dumpdyn_main(command.split())

    def test_sumpds(self):
        """Test the sum of pdss."""
        hen.sum_fspec.main([
            os.path.join(self.datadir,
                         'monol_testA_E3-50_pds') + HEN_FILE_EXTENSION,
            os.path.join(self.datadir,
                         'monol_testB_E3-50_pds') + HEN_FILE_EXTENSION,
            '-o', os.path.join(self.datadir,
                               'monol_test_sum' + HEN_FILE_EXTENSION)])

    # def test_dumpdyncpds(self):
    #     """Test dumping CPDS file."""
    #     command = '--noplot ' + \
    #         os.path.join(self.datadir,
    #                      'monol_test_E3-50_cpds') + \
    #         HEN_FILE_EXTENSION
    #     hen.fspec.dumpdyn_main(command.split())

    def test_rebinlc(self):
        """Test LC rebinning."""
        command = '{0} -r 4'.format(
            os.path.join(self.datadir, 'monol_testA_E3-50_lc') +
            HEN_FILE_EXTENSION)
        hen.rebin.main(command.split())

    def test_rebinpds(self):
        """Test PDS rebinning 1."""
        command = '{0} -r 2'.format(
            os.path.join(self.datadir, 'monol_testA_E3-50_pds') +
            HEN_FILE_EXTENSION)
        hen.rebin.main(command.split())
        os.path.exists(os.path.join(self.datadir,
                                    'monol_testA_E3-50_pds_rebin2' +
                                    HEN_FILE_EXTENSION))

    def test_rebinpds_geom(self):
        """Test geometrical PDS rebinning."""
        command = '{0} {1} -r 1.03'.format(
            os.path.join(self.datadir, 'monol_testA_E3-50_pds') +
            HEN_FILE_EXTENSION,
            os.path.join(self.datadir, 'monol_testB_E3-50_pds') +
            HEN_FILE_EXTENSION
            )
        hen.rebin.main(command.split())
        os.path.exists(os.path.join(self.datadir,
                                    'monol_testA_E3-50_pds_rebin1.03' +
                                    HEN_FILE_EXTENSION))
        os.path.exists(os.path.join(self.datadir,
                                    'monol_testB_E3-50_pds_rebin1.03' +
                                    HEN_FILE_EXTENSION))

    def test_rebincpds(self):
        """Test CPDS rebinning."""
        command = '{0} -r 2'.format(
            os.path.join(self.datadir, 'monol_test_E3-50_cpds') +
            HEN_FILE_EXTENSION)
        hen.rebin.main(command.split())
        os.path.exists(os.path.join(self.datadir,
                                    'monol_test_E3-50_cpds_rebin2' +
                                    HEN_FILE_EXTENSION))

    def test_rebincpds_geom(self):
        """Test CPDS geometrical rebinning."""
        command = '{0} -r 1.03'.format(
            os.path.join(self.datadir, 'monol_test_E3-50_cpds') +
            HEN_FILE_EXTENSION)
        hen.rebin.main(command.split())
        os.path.exists(os.path.join(self.datadir,
                                    'monol_test_E3-50_cpds_rebin1.03' +
                                    HEN_FILE_EXTENSION))

    def test_save_lags(self):
        fname = os.path.join(self.datadir,
                             'monol_test_E3-50_cpds_rebin2' +
                             HEN_FILE_EXTENSION)
        hen.timelags.main([fname])
        out = hen.base.hen_root(fname) + '_lags.qdp'
        os.path.exists(out)

    def test_save_fvar(self):
        fname = os.path.join(self.datadir,
                             'monol_testA_E3-50_lc' + HEN_FILE_EXTENSION)
        hen.exvar.main([fname, "-c", "10", "--fraction-step", "0.6",
                        "--norm", "fvar"])
        out = hen.base.hen_root(fname) + "_fvar" + '.qdp'
        os.path.exists(out)

    def test_save_excvar(self):
        fname = os.path.join(self.datadir,
                             'monol_testA_E3-50_lc' +
                             HEN_FILE_EXTENSION)
        hen.exvar.main([fname])
        out = hen.base.hen_root(fname) + "_excvar" + '.qdp'
        os.path.exists(out)

    def test_save_excvar_norm(self):
        fname = os.path.join(self.datadir,
                             'monol_testA_E3-50_lc' +
                             HEN_FILE_EXTENSION)
        hen.exvar.main([fname, "--norm", "norm_excvar"])
        out = hen.base.hen_root(fname) + "_norm_excvar" + '.qdp'
        os.path.exists(out)

    def test_save_excvar_wrong_norm(self):
        fname = os.path.join(self.datadir,
                             'monol_testA_E3-50_lc' +
                             HEN_FILE_EXTENSION)
        with pytest.raises(ValueError) as excinfo:
            hen.exvar.main([fname, '--norm', 'cicciput'])
        assert 'Normalization must be fvar, ' in str(excinfo.value)

    def test_fit_pds(self):
        modelstring = '''
from astropy.modeling import models
model = models.Const1D()
        '''
        modelfile = 'bubu__model__.py'
        with open(modelfile, 'w') as fobj:
            print(modelstring, file=fobj)
        pdsfile1 = \
            os.path.join(self.datadir,
                         'monol_testA_E3-50_pds' + HEN_FILE_EXTENSION)
        pdsfile2 = \
            os.path.join(self.datadir,
                         'monol_testB_E3-50_pds' + HEN_FILE_EXTENSION)

        command = '{0} {1} -m {2} --frequency-interval 0 10'.format(
            pdsfile1,
            pdsfile2,
            modelfile)
        hen.modeling.main_model(command.split())

        out0 = os.path.join(self.datadir,
                            'monol_testA_E3-50_pds_bestfit.p')
        out1 = os.path.join(self.datadir,
                            'monol_testB_E3-50_pds_bestfit.p')
        assert os.path.exists(out0)
        assert os.path.exists(out1)
        m, k, c = hen.io.load_model(
            os.path.join(self.datadir,
                         'monol_testB_E3-50_pds_bestfit.p'))
        assert hasattr(m, 'amplitude')
        os.unlink(out0)
        os.unlink(out1)

        out0 = os.path.join(self.datadir,
                            'monol_testA_E3-50_pds_fit' + HEN_FILE_EXTENSION)
        out1 = os.path.join(self.datadir,
                            'monol_testB_E3-50_pds_fit' + HEN_FILE_EXTENSION)
        assert os.path.exists(out0)
        assert os.path.exists(out1)
        spec = hen.io.load_pds(out0)
        assert hasattr(spec, 'best_fits')

    def test_fit_cpds(self):
        modelstring = '''
from astropy.modeling import models
model = models.Const1D()
        '''
        modelfile = 'bubu__model__.py'
        with open(modelfile, 'w') as fobj:
            print(modelstring, file=fobj)
        pdsfile1 = \
            os.path.join(self.datadir,
                         'monol_test_E3-50_cpds' + HEN_FILE_EXTENSION)

        command = '{0} -m {1} --frequency-interval 0 10'.format(
            pdsfile1,
            modelfile)
        hen.modeling.main_model(command.split())

        out0 = os.path.join(self.datadir,
                            'monol_test_E3-50_cpds_bestfit.p')
        assert os.path.exists(out0)
        m, k, c = hen.io.load_model(out0)
        assert hasattr(m, 'amplitude')
        os.unlink(out0)

        out0 = \
            os.path.join(self.datadir,
                         'monol_test_E3-50_cpds_fit' + HEN_FILE_EXTENSION)
        assert os.path.exists(out0)
        spec = hen.io.load_pds(out0)
        assert hasattr(spec, 'best_fits')

    def test_fit_pds_f_no_of_intervals_invalid(self):
        modelstring = '''
from astropy.modeling import models
model = models.Const1D()
        '''
        modelfile = 'bubu__model__.py'
        with open(modelfile, 'w') as fobj:
            print(modelstring, file=fobj)
        pdsfile1 = \
            os.path.join(self.datadir,
                         'monol_testA_E3-50_pds' + HEN_FILE_EXTENSION)
        pdsfile2 = \
            os.path.join(self.datadir,
                         'monol_testB_E3-50_pds' + HEN_FILE_EXTENSION)

        command = '{0} {1} -m {2} --frequency-interval 0 1 9'.format(pdsfile1,
                                                                     pdsfile2,
                                                                     modelfile)
        with pytest.raises(ValueError) as excinfo:
            hen.modeling.main_model(command.split())

        assert "Invalid number of frequencies specified" in str(excinfo.value)

    # def test_dumpdyncpds_reb(self):
    #     """Test dumping rebinned CPDS file."""
    #     command = '--noplot ' + \
    #         os.path.join(self.datadir,
    #                      'monol_test_E3-50_cpds_rebin1.03') + \
    #         HEN_FILE_EXTENSION
    #     hen.fspec.dumpdyn_main(command.split())

    def test_savexspec(self):
        """Test save as Xspec 1."""
        command = '{0}'.format(
            os.path.join(self.datadir, 'monol_testA_E3-50_pds_rebin2') +
            HEN_FILE_EXTENSION)
        hen.save_as_xspec.main(command.split())
        os.path.exists(os.path.join(self.datadir,
                                    'monol_testA_E3-50_pds_rebin2.pha'))

    def test_savexspec_geom(self):
        """Test save as Xspec 2."""
        command = '{0}'.format(
            os.path.join(self.datadir, 'monol_test_E3-50_cpds_rebin1.03') +
            HEN_FILE_EXTENSION)
        hen.save_as_xspec.main(command.split())

        os.path.exists(os.path.join(self.datadir,
                                    'monol_test_E3-50_cpds_rebin1.03.pha'))
        os.path.exists(
            os.path.join(self.datadir,
                         'monol_test_E3-50_cpds_rebin1.03_lags.pha'))

    def test_create_gti(self):
        """Test creating a GTI file."""
        fname = os.path.join(self.datadir, 'monol_testA_E3-50_lc_rebin4') + \
            HEN_FILE_EXTENSION
        command = "{0} -f counts>0 -c --debug".format(fname)
        hen.create_gti.main(command.split())

    def test_apply_gti(self):
        """Test applying a GTI file."""
        fname = os.path.join(self.datadir, 'monol_testA_E3-50_rebin4_gti') + \
            HEN_FILE_EXTENSION
        lcfname = os.path.join(self.datadir, 'monol_testA_E3-50_lc') + \
            HEN_FILE_EXTENSION
        lcoutname = os.path.join(self.datadir,
                                 'monol_testA_E3-50_lc_gtifilt') + \
            HEN_FILE_EXTENSION
        command = "{0} -a {1} --debug".format(lcfname, fname)
        hen.create_gti.main(command.split())
        hen.io.load_lcurve(lcoutname)

    def test_create_gti_and_minlen(self):
        """Test creating a GTI file and apply minimum length."""
        fname = os.path.join(self.datadir, 'monol_testA_E3-50_lc_rebin4') + \
            HEN_FILE_EXTENSION
        command = "{0} -f counts>0 -c -l 10 --debug".format(fname)
        hen.create_gti.main(command.split())

    def test_create_gti_and_apply(self):
        """Test applying a GTI file and apply minimum length."""
        fname = os.path.join(self.datadir, 'monol_testA_E3-50_rebin4_gti') + \
            HEN_FILE_EXTENSION
        lcfname = os.path.join(self.datadir, 'monol_testA_E3-50_lc') + \
            HEN_FILE_EXTENSION
        command = "{0} -a {1} -l 10 --debug".format(lcfname, fname)
        hen.create_gti.main(command.split())

    def test_readfile(self):
        """Test reading and dumping a HENDRICS file."""
        fname = os.path.join(self.datadir, 'monol_testA_E3-50_rebin4_gti') + \
            HEN_FILE_EXTENSION
        command = "{0}".format(fname)

        hen.io.main(command.split())

    def test_readfile_fits(self):
        """Test reading and dumping a FITS file."""
        fitsname = os.path.join(self.datadir, 'monol_testA.evt')
        command = "{0}".format(fitsname)

        hen.io.main(command.split())

    def test_save_as_qdp(self):
        """Test saving arrays in a qdp file."""
        arrays = [np.array([0, 1, 3]), np.array([1, 4, 5])]
        errors = [np.array([1, 1, 1]), np.array([[1, 0.5], [1, 0.5], [1, 1]])]
        hen.io.save_as_qdp(arrays, errors,
                           filename=os.path.join(self.datadir,
                                                 "monol_test_qdp.txt"))
        hen.io.save_as_qdp(arrays, errors,
                           filename=os.path.join(self.datadir,
                                                 "monol_test_qdp.txt"),
                           mode='a')

    def test_save_as_ascii(self):
        """Test saving arrays in a ascii file."""
        array = np.array([0, 1, 3])
        errors = np.array([1, 1, 1])
        hen.io.save_as_ascii(
            [array, errors],
            filename=os.path.join(self.datadir, "monol_test.txt"),
            colnames=["array", "err"])

    def test_get_file_type(self):
        """Test getting file type."""
        file_list = {'events': 'monol_testA_nustar_fpma_ev',
                     'lc': 'monol_testA_E3-50_lc',
                     'pds': 'monol_testA_E3-50_pds',
                     'gti': 'monol_testA_E3-50_rebin4_gti',
                     'cpds': 'monol_test_E3-50_cpds'}
        for realtype in file_list.keys():
            fname = os.path.join(self.datadir,
                                 file_list[realtype] + HEN_FILE_EXTENSION)
            ftype, _ = hen.io.get_file_type(fname)
            assert ftype == realtype, "File types do not match"

    def test_exposure(self):
        """Test exposure calculations from unfiltered files."""
        lcname = os.path.join(self.datadir,
                              'monol_testA_E3-50_lc' + HEN_FILE_EXTENSION)
        ufname = os.path.join(self.datadir, 'monol_testA_uf.evt')
        command = "{0} {1}".format(lcname, ufname)

        hen.exposure.main(command.split())
        fname = os.path.join(self.datadir,
                             'monol_testA_E3-50_lccorr' + HEN_FILE_EXTENSION)
        assert os.path.exists(fname)
        ftype, contents = hen.io.get_file_type(fname)

        assert isinstance(contents, Lightcurve)
        assert hasattr(contents, 'expo')

    def test_plot_lin(self):
        """Test plotting with linear axes."""
        pname = os.path.join(self.datadir, 'monol_testA_E3-50_pds_fit') + \
            HEN_FILE_EXTENSION
        cname = os.path.join(self.datadir, 'monol_test_E3-50_cpds_fit') + \
            HEN_FILE_EXTENSION
        lname = os.path.join(self.datadir, 'monol_testA_E3-50_lc') + \
            HEN_FILE_EXTENSION
        hen.plot.main([pname, cname, lname, '--noplot', '--xlin', '--ylin',
                      '-o', 'dummy.qdp'])
        hen.plot.main([lname, '--noplot',
                       '--axes', 'time', 'counts', '--xlin', '--ylin',
                       '-o', 'dummy.qdp'])

    def test_plot_lcurve_baseline(self):
        a_in = os.path.join(self.datadir,
                            'monol_testA_E3-50_lc' + HEN_FILE_EXTENSION)
        base_file = hen.base.hen_root(a_in) + '_lc_baseline' + \
            HEN_FILE_EXTENSION
        hen.plot.main([base_file, '--noplot', '-o', 'dummy_base.qdp'])
        filedata = np.genfromtxt('dummy_base.qdp')

        assert filedata.shape[1] == 3

    def test_plot_log(self):
        """Test plotting with log axes."""
        pname = os.path.join(
            self.datadir,
            'monol_testA_E3-50_pds_rebin1.03' + HEN_FILE_EXTENSION)
        cname = os.path.join(
            self.datadir,
            'monol_test_E3-50_cpds_rebin1.03' + HEN_FILE_EXTENSION)
        hen.plot.main([pname, cname, '--noplot', '--xlog', '--ylog',
                       '-o', 'dummy.qdp'])
        hen.plot.main([pname, '--noplot', '--axes', 'power', 'power_err',
                       '--xlin', '--ylin',
                       '-o', 'dummy.qdp'])

    def test_plot_save_figure(self):
        """Test plotting and saving figure."""
        pname = os.path.join(
            self.datadir,
            'monol_testA_E3-50_pds_rebin1.03' + HEN_FILE_EXTENSION)
        hen.plot.main([pname, '--noplot', '--figname',
                      os.path.join(self.datadir,
                                   'monol_testA_E3-50_pds_rebin1.03.png'),
                      '-o', 'dummy.qdp'])

    def test_plot_color(self):
        """Test plotting with linear axes."""
        lname = os.path.join(self.datadir,
                             'monol_testA_nustar_fpma_E_10-5_over_5-3') + \
            HEN_FILE_EXTENSION
        cname = os.path.join(self.datadir,
                             'monol_testA_nustar_fpma_E_10-5_over_5-3') + \
            HEN_FILE_EXTENSION
        hen.plot.main([cname, lname, '--noplot', '--xlog', '--ylog', '--CCD',
                      '-o', 'dummy.qdp'])

    def test_plot_hid(self):
        """Test plotting with linear axes."""
        # also produce a light curve with the same binning
        command = ('{0} -b 100 --e-interval {1} {2}').format(
            os.path.join(self.datadir, 'monol_testA_nustar_fpma_ev_calib' +
                         HEN_FILE_EXTENSION), 3, 10)

        hen.lcurve.main(command.split())
        lname = os.path.join(self.datadir,
                             'monol_testA_nustar_fpma_E3-10_lc') + \
            HEN_FILE_EXTENSION
        os.path.exists(lname)
        cname = os.path.join(self.datadir,
                             'monol_testA_nustar_fpma_E_10-5_over_5-3') + \
            HEN_FILE_EXTENSION
        hen.plot.main([cname, lname, '--noplot', '--xlog', '--ylog', '--HID',
                      '-o', 'dummy.qdp'])

    @classmethod
    def teardown_class(self):
        """Test a full run of the scripts (command lines)."""

        file_list = \
            glob.glob(os.path.join(self.datadir,
                                   '*monol_test*') + HEN_FILE_EXTENSION) + \
            glob.glob(os.path.join(self.datadir,
                                   '*lcurve*') + HEN_FILE_EXTENSION) + \
            glob.glob(os.path.join(self.datadir,
                                   '*lcurve*.txt')) + \
            glob.glob(os.path.join(self.datadir,
                                   '*.log')) + \
            glob.glob(os.path.join(self.datadir,
                                   '*monol_test*.dat')) + \
            glob.glob(os.path.join(self.datadir,
                                   '*monol_test*.png')) + \
            glob.glob(os.path.join(self.datadir,
                                   '*monol_test*.txt')) + \
            glob.glob(os.path.join(self.datadir,
                                   'monol_test_fake*.evt')) + \
            glob.glob(os.path.join(self.datadir,
                                   'bubu*'))
        for f in file_list:
            print("Removing " + f)
            os.remove(f)
